from __future__ import absolute_import, division, print_function

import pytest

sa = pytest.importorskip('sqlalchemy')

import itertools
import sqlite3
from distutils.version import LooseVersion


import datashape
from odo import into, discover
import numpy as np
import pandas as pd
import pandas.util.testing as tm
from pandas import DataFrame
from toolz import unique

from odo import odo, drop, resource
from blaze import data as bz_data
from blaze.compatibility import xfail
from blaze.compute.sql import compute, select, lower_column, compute_up
from blaze.expr import (
    by,
    coalesce,
    cos,
    count,
    datetime as bz_datetime,
    exp,
    floor,
    greatest,
    join,
    least,
    mean,
    merge,
    nunique,
    sin,
    sum,
    summary,
    symbol,
    transform,
)
from blaze.utils import tmpfile, example, normalize, literal_compile


def computefull(t, s):
    return select(compute(t, s, return_type='native'))


names = ('tbl%d' % i for i in itertools.count())


@pytest.fixture(scope='module')
def city_data():
    # make the engine
    engine = sa.create_engine('sqlite:///:memory:')
    metadata = sa.MetaData(engine)

    # name table
    name = sa.Table('name', metadata,
                    sa.Column('id', sa.Integer),
                    sa.Column('name', sa.String),
                    )
    name.create()

    # city table
    city = sa.Table('city', metadata,
                    sa.Column('id', sa.Integer),
                    sa.Column('city', sa.String),
                    sa.Column('country', sa.String),
                    )
    city.create()

    s = symbol('s', discover(engine))
    return {'engine': engine, 'metadata': metadata, 'name': name, 'city': city,
            's': s}


t = symbol('t', 'var * {name: string, amount: int, id: int}')
t_str_cat = symbol('t',
                   """var * {name: string,
                             amount: int,
                             id: int,
                             comment: string,
                             product: string}""")

nt = symbol('t', 'var * {name: ?string, amount: float64, id: int}')

metadata = sa.MetaData()

s = sa.Table('accounts', metadata,
             sa.Column('name', sa.String),
             sa.Column('amount', sa.Integer),
             sa.Column('id', sa.Integer, primary_key=True))

s_str_cat = sa.Table('accounts2', metadata,
                     sa.Column('name', sa.String),
                     sa.Column('amount', sa.Integer),
                     sa.Column('id', sa.Integer, primary_key=True),
                     sa.Column('comment', sa.String),
                     sa.Column('product', sa.String))

tdate = symbol('t',
               """var * {
                    name: string,
                    amount: int,
                    id: int,
                    occurred_on: datetime
                }""")

ns = sa.Table('nullaccounts', metadata,
              sa.Column('name', sa.String, nullable=True),
              sa.Column('amount', sa.REAL),
              sa.Column('id', sa.Integer, primary_key=True),
              )

sdate = sa.Table('accdate', metadata,
                 sa.Column('name', sa.String),
                 sa.Column('amount', sa.Integer),
                 sa.Column('id', sa.Integer, primary_key=True),
                 sa.Column('occurred_on', sa.DateTime))


tbig = symbol('tbig',
              'var * {name: string, sex: string[1], amount: int, id: int}')

sbig = sa.Table('accountsbig', metadata,
                sa.Column('name', sa.String),
                sa.Column('sex', sa.String),
                sa.Column('amount', sa.Integer),
                sa.Column('id', sa.Integer, primary_key=True))


def test_table():
    result = str(computefull(t, s))
    expected = """
    SELECT accounts.name, accounts.amount, accounts.id
    FROM accounts
    """.strip()

    assert normalize(result) == normalize(expected)


def test_projection():
    print(compute(t[['name', 'amount']], s, return_type='native'))
    assert str(compute(t[['name', 'amount']], s, return_type='native')) == \
        str(sa.select([s.c.name, s.c.amount]))


def test_eq():
    assert str(compute(t['amount'] == 100, s, post_compute=False, return_type='native')) == \
        str(s.c.amount == 100)


def test_eq_unicode():
    assert str(compute(t['name'] == u'Alice', s, post_compute=False, return_type='native')) == \
        str(s.c.name == u'Alice')


def test_selection():
    assert str(compute(t[t['amount'] == 0], s, return_type='native')) == \
        str(sa.select([s]).where(s.c.amount == 0))
    assert str(compute(t[t['amount'] > 150], s, return_type='native')) == \
        str(sa.select([s]).where(s.c.amount > 150))


def test_arithmetic():
    assert str(compute(t['amount'] + t['id'], s, return_type='native')) == \
        str(sa.select([s.c.amount + s.c.id]))
    assert str(compute(t['amount'] + t['id'], s, post_compute=False, return_type='native')) == \
        str(s.c.amount + s.c.id)
    assert str(compute(t['amount'] * t['id'], s, post_compute=False, return_type='native')) == \
        str(s.c.amount * s.c.id)

    assert str(compute(t['amount'] * 2, s, post_compute=False, return_type='native')) == \
        str(s.c.amount * 2)
    assert str(compute(2 * t['amount'], s, post_compute=False, return_type='native')) == \
        str(2 * s.c.amount)

    assert (str(compute(~(t['amount'] > 10), s, post_compute=False, return_type='native')) ==
            "accounts.amount <= :amount_1")

    assert str(compute(t['amount'] + t['id'] * 2, s, return_type='native')) == \
        str(sa.select([s.c.amount + s.c.id * 2]))


def test_join():
    metadata = sa.MetaData()
    lhs = sa.Table('amounts', metadata,
                   sa.Column('name', sa.String),
                   sa.Column('amount', sa.Integer))

    rhs = sa.Table('ids', metadata,
                   sa.Column('name', sa.String),
                   sa.Column('id', sa.Integer))

    expected = lhs.join(rhs, lhs.c.name == rhs.c.name)
    expected = select(list(unique(expected.columns, key=lambda c:
                                  c.name))).select_from(expected)

    L = symbol('L', 'var * {name: string, amount: int}')
    R = symbol('R', 'var * {name: string, id: int}')
    joined = join(L, R, 'name')

    result = compute(joined, {L: lhs, R: rhs}, return_type='native')

    assert normalize(str(result)) == normalize("""
    SELECT amounts.name, amounts.amount, ids.id
    FROM amounts JOIN ids ON amounts.name = ids.name""")

    assert str(select(result)) == str(select(expected))

    # Schemas match
    assert list(result.c.keys()) == list(joined.fields)

    # test sort on join

    result = compute(joined.sort('amount'), {L: lhs, R: rhs}, return_type='native')
    assert normalize(str(result)) == normalize("""
     select
        anon_1.name,
        anon_1.amount,
        anon_1.id
    from (select
              amounts.name as name,
              amounts.amount as amount,
              ids.id as id
          from
              amounts
          join
              ids
          on
              amounts.name = ids.name) as anon_1
    order by
        anon_1.amount asc""")


def test_clean_complex_join():
    metadata = sa.MetaData()
    lhs = sa.Table('amounts', metadata,
                   sa.Column('name', sa.String),
                   sa.Column('amount', sa.Integer))

    rhs = sa.Table('ids', metadata,
                   sa.Column('name', sa.String),
                   sa.Column('id', sa.Integer))

    L = symbol('L', 'var * {name: string, amount: int}')
    R = symbol('R', 'var * {name: string, id: int}')

    joined = join(L[L.amount > 0], R, 'name')

    result = compute(joined, {L: lhs, R: rhs}, return_type='native')

    expected1 = """
        SELECT amounts.name, amounts.amount, ids.id
        FROM amounts JOIN ids ON amounts.name = ids.name
        WHERE amounts.amount > :amount_1"""

    expected2 = """
        SELECT alias.name, alias.amount, ids.id
        FROM (SELECT amounts.name AS name, amounts.amount AS amount
              FROM amounts
              WHERE amounts.amount > :amount_1) AS alias
        JOIN ids ON alias.name = ids.name"""

    assert (normalize(str(result)) == normalize(expected1) or
            normalize(str(result)) == normalize(expected2))


def test_multi_column_join():
    metadata = sa.MetaData()
    lhs = sa.Table('aaa', metadata,
                   sa.Column('x', sa.Integer),
                   sa.Column('y', sa.Integer),
                   sa.Column('z', sa.Integer))

    rhs = sa.Table('bbb', metadata,
                   sa.Column('w', sa.Integer),
                   sa.Column('x', sa.Integer),
                   sa.Column('y', sa.Integer))

    L = symbol('L', 'var * {x: int, y: int, z: int}')
    R = symbol('R', 'var * {w: int, x: int, y: int}')
    joined = join(L, R, ['x', 'y'])

    expected = lhs.join(rhs, (lhs.c.x == rhs.c.x)
                        & (lhs.c.y == rhs.c.y))
    expected = select(list(unique(expected.columns, key=lambda c:
                                  c.name))).select_from(expected)

    result = compute(joined, {L: lhs, R: rhs}, return_type='native')

    assert str(result) == str(expected)

    assert str(select(result)) == str(select(expected))

    # Schemas match
    print(result.c.keys())
    print(joined.fields)
    assert list(result.c.keys()) == list(joined.fields)


def test_unary_op():
    assert str(compute(exp(t['amount']), s, post_compute=False, return_type='native')) == \
        str(sa.func.exp(s.c.amount))

    assert str(compute(-t['amount'], s, post_compute=False, return_type='native')) == \
        str(-s.c.amount)


@pytest.mark.parametrize('unbiased', [True, False])
def test_std(unbiased):
    assert str(compute(t.amount.std(unbiased=unbiased), s, post_compute=False, return_type='native')) == \
        str(getattr(sa.func,
                    'stddev_%s' % ('samp' if unbiased else 'pop'))(s.c.amount))


@pytest.mark.parametrize('unbiased', [True, False])
def test_var(unbiased):
    assert str(compute(t.amount.var(unbiased=unbiased), s, post_compute=False, return_type='native')) == \
        str(getattr(sa.func,
                    'var_%s' % ('samp' if unbiased else 'pop'))(s.c.amount))


def test_reductions():
    assert str(compute(sum(t['amount']), s, post_compute=False, return_type='native')) == \
        str(sa.sql.functions.sum(s.c.amount))
    assert str(compute(mean(t['amount']), s, post_compute=False, return_type='native')) == \
        str(sa.sql.func.avg(s.c.amount))
    assert str(compute(count(t['amount']), s, post_compute=False, return_type='native')) == \
        str(sa.sql.func.count(s.c.amount))

    assert 'amount_sum' == compute(
        sum(t['amount']), s, post_compute=False, return_type='native').name


def test_reduction_with_invalid_axis_argument():
    with pytest.raises(ValueError):
        compute(t.amount.mean(axis=1))

    with pytest.raises(ValueError):
        compute(t.count(axis=1))

    with pytest.raises(ValueError):
        compute(t[['amount', 'id']].count(axis=1))


def test_nelements():
    rhs = str(compute(t.count(), s, return_type='native'))
    assert str(compute(t.nelements(), s, return_type='native')) == rhs
    assert str(compute(t.nelements(axis=None), s, return_type='native')) == rhs
    assert str(compute(t.nelements(axis=0), s, return_type='native')) == rhs
    assert str(compute(t.nelements(axis=(0,)), s, return_type='native')) == rhs


@pytest.mark.xfail(raises=Exception, reason="We don't support axis=1 for"
                   " Record datashapes")
def test_nelements_axis_1():
    assert compute(t.nelements(axis=1), s, return_type='native') == len(s.columns)


def test_count_on_table():
    result = compute(t.count(), s, return_type='native')
    assert normalize(str(result)) == normalize("""
    SELECT count(accounts.id) as count_1
    FROM accounts""")

    result = compute(t[t.amount > 0].count(), s, return_type='native')
    assert (
        normalize(str(result)) == normalize("""
        SELECT count(accounts.id) as t_count
        FROM accounts
        WHERE accounts.amount > :amount_1""")

        or

        normalize(str(result)) == normalize("""
        SELECT count(alias.id) as t_count
        FROM (SELECT accounts.name AS name, accounts.amount AS amount, accounts.id AS id
              FROM accounts
              WHERE accounts.amount > :amount_1) as alias_2"""))


def test_distinct():
    result = str(compute(t['amount'].distinct(), s, post_compute=False, return_type='native'))

    assert 'distinct' in result.lower()
    assert 'amount' in result.lower()

    print(result)
    assert result == str(sa.distinct(s.c.amount))


def test_distinct_multiple_columns():
    assert normalize(str(compute(t.distinct(), s, return_type='native'))) == normalize("""
    SELECT DISTINCT accounts.name, accounts.amount, accounts.id
    FROM accounts""")


def test_nunique():
    result = str(computefull(nunique(t['amount']), s))

    print(result)
    assert 'distinct' in result.lower()
    assert 'count' in result.lower()
    assert 'amount' in result.lower()


def test_nunique_table():
    result = normalize(str(computefull(t.nunique(), s)))
    expected = normalize("""SELECT count(alias.id) AS tbl_row_count
FROM (SELECT DISTINCT accounts.name AS name, accounts.amount AS amount, accounts.id AS id
FROM accounts) as alias""")
    assert result == expected


@xfail(reason="Fails because SQLAlchemy doesn't seem to know binary reductions")
def test_binary_reductions():
    assert str(compute(any(t['amount'] > 150), s, return_type='native')) == \
        str(sa.sql.functions.any(s.c.amount > 150))


def test_by():
    expr = by(t['name'], total=t['amount'].sum())
    result = compute(expr, s, return_type='native')
    expected = sa.select([s.c.name,
                          sa.sql.functions.sum(s.c.amount).label('total')]
                         ).group_by(s.c.name)

    assert str(result) == str(expected)


def test_by_head():
    t2 = t.head(100)
    expr = by(t2['name'], total=t2['amount'].sum())
    result = compute(expr, s, return_type='native')
    expected = """
    SELECT accounts.name, sum(accounts.amount) as total
    FROM accounts
    GROUP by accounts.name
    LIMIT :param_1"""

    assert normalize(str(result)) == normalize(str(expected))


def test_by_two():
    expr = by(tbig[['name', 'sex']], total=tbig['amount'].sum())
    result = compute(expr, sbig, return_type='native')
    expected = (sa.select([sbig.c.name,
                           sbig.c.sex,
                           sa.sql.functions.sum(sbig.c.amount).label('total')])
                .group_by(sbig.c.name, sbig.c.sex))

    assert str(result) == str(expected)


def test_by_three():
    result = compute(by(tbig[['name', 'sex']],
                        total=(tbig['id'] + tbig['amount']).sum()),
                     sbig, return_type='native')

    assert normalize(str(result)) == normalize("""
    SELECT accountsbig.name,
           accountsbig.sex,
           sum(accountsbig.id + accountsbig.amount) AS total
    FROM accountsbig GROUP BY accountsbig.name, accountsbig.sex
    """)


def test_by_summary_clean():
    expr = by(t.name, min=t.amount.min(), max=t.amount.max())
    result = compute(expr, s, return_type='native')

    expected = """
    SELECT accounts.name, max(accounts.amount) AS max, min(accounts.amount) AS min
    FROM accounts
    GROUP BY accounts.name
    """

    assert normalize(str(result)) == normalize(expected)


def test_by_summary_single_column():
    expr = by(t.name, n=t.name.count(), biggest=t.name.max())
    result = compute(expr, s, return_type='native')

    expected = """
    SELECT accounts.name, max(accounts.name) AS biggest, count(accounts.name) AS n
    FROM accounts
    GROUP BY accounts.name
    """

    assert normalize(str(result)) == normalize(expected)


def test_join_projection():
    metadata = sa.MetaData()
    lhs = sa.Table('amounts', metadata,
                   sa.Column('name', sa.String),
                   sa.Column('amount', sa.Integer))

    rhs = sa.Table('ids', metadata,
                   sa.Column('name', sa.String),
                   sa.Column('id', sa.Integer))

    L = symbol('L', 'var * {name: string, amount: int}')
    R = symbol('R', 'var * {name: string, id: int}')
    want = join(L, R, 'name')[['amount', 'id']]

    result = compute(want, {L: lhs, R: rhs}, return_type='native')
    print(result)
    assert 'join' in str(result).lower()
    assert result.c.keys() == ['amount', 'id']
    assert 'amounts.name = ids.name' in str(result)


def test_sort():
    assert str(compute(t.sort('amount'), s, return_type='native')) == \
        str(select(s).order_by(sa.asc(s.c.amount)))

    assert str(compute(t.sort('amount', ascending=False), s, return_type='native')) == \
        str(select(s).order_by(sa.desc(s.c.amount)))


def test_multicolumn_sort():
    assert str(compute(t.sort(['amount', 'id']), s, return_type='native')) == \
        str(select(s).order_by(sa.asc(s.c.amount), sa.asc(s.c.id)))

    assert str(compute(t.sort(['amount', 'id'], ascending=False), s, return_type='native')) == \
        str(select(s).order_by(sa.desc(s.c.amount), sa.desc(s.c.id)))


def test_sort_on_distinct():
    assert normalize(str(compute(t.amount.sort(), s, return_type='native'))) == normalize("""
            SELECT accounts.amount
            FROM accounts
            ORDER BY accounts.amount ASC""")

    assert normalize(str(compute(t.amount.distinct().sort(), s, return_type='native'))) == normalize("""
            SELECT DISTINCT accounts.amount as amount
            FROM accounts
            ORDER BY amount ASC""")


def test_head():
    assert str(compute(t.head(2), s, return_type='native')) == str(select(s).limit(2))


def test_sample():
    order_by = select(s).order_by(sa.func.random())
    assert str(compute(t.sample(n=5), s)) == str(order_by.limit(5))


def test_label():
    assert (str(compute((t['amount'] * 10).label('foo'),
                        s, post_compute=False, return_type='native')) ==
            str((s.c.amount * 10).label('foo')))


def test_relabel_table():
    result = compute(t.relabel(name='NAME', id='ID'), s, return_type='native')
    expected = select([
        s.c.name.label('NAME'),
        s.c.amount,
        s.c.id.label('ID'),
    ])

    assert str(result) == str(expected)


def test_relabel_columns_over_selection():
    result = compute(
        t[t['amount'] > 150].relabel(name='NAME'), s, return_type='native'
    )
    expected = sa.select(
        [s.c.name.label('NAME'), s.c.amount, s.c.id]
    ).where(s.c.amount > 150)
    assert literal_compile(result) == literal_compile(expected)


def test_relabel_projection():
    result = compute(
        t[['name', 'id']].relabel(name='new_name', id='new_id'),
        s,
        return_type='native'
    )
    assert normalize(str(result)) == normalize(
        """SELECT
            accounts.name AS new_name,
            accounts.id AS new_id
        FROM accounts""",
    )


def test_relabel_join():
    metadata = sa.MetaData()
    amounts_tbl = sa.Table('amounts', metadata,
                           sa.Column('id', sa.Integer),
                           sa.Column('amount', sa.Integer))
    ids_tbl = sa.Table('ids', metadata,
                       sa.Column('name', sa.String),
                       sa.Column('id', sa.Integer))
    joined_sql = amounts_tbl.join(ids_tbl, amounts_tbl.c.id == ids_tbl.c.id)

    amounts = symbol('amounts', discover(amounts_tbl))
    ids = symbol('ids', discover(ids_tbl))
    joined_expr = join(amounts, ids, 'id').relabel({'amount': 'balance'})

    ns = {amounts: amounts_tbl, ids: ids_tbl}
    result = str(compute(joined_expr, ns, return_type='native'))
    expected = sa.select([amounts_tbl.c.id, amounts_tbl.c.amount.label('balance'),
                          ids_tbl.c.name], from_obj=joined_sql)
    assert result == literal_compile(expected)



def test_merge():
    col = (t['amount'] * 2).label('new')

    expr = merge(t['name'], col)

    result = str(compute(expr, s, return_type='native'))

    assert 'amount * ' in result
    assert 'FROM accounts' in result
    assert 'SELECT accounts.name' in result
    assert 'new' in result


def test_projection_of_selection():
    print(compute(t[t['amount'] < 0][['name', 'amount']], s, return_type='native'))
    assert len(str(compute(t[t['amount'] < 0], s, return_type='native'))) > \
        len(str(compute(t[t['amount'] < 0][['name', 'amount']], s, return_type='native')))


def test_outer_join():
    L = symbol('L', 'var * {id: int, name: string, amount: real}')
    R = symbol('R', 'var * {city: string, id: int}')

    with tmpfile('db') as fn:
        uri = 'sqlite:///' + fn
        engine = bz_data(uri).data

        _left = [(1, 'Alice', 100),
                 (2, 'Bob', 200),
                 (4, 'Dennis', 400)]

        left = bz_data(uri + '::left', dshape=L.dshape)
        into(left, _left)

        _right = [('NYC', 1),
                  ('Boston', 1),
                  ('LA', 3),
                  ('Moscow', 4)]
        right = bz_data(uri + '::right', dshape=R.dshape)
        into(right, _right)

        conn = engine.connect()

        query = compute(join(L, R, how='inner'),
                        {L: left, R: right},
                        post_compute=False, return_type='native')
        result = list(map(tuple, conn.execute(query).fetchall()))

        assert set(result) == set(
            [(1, 'Alice', 100, 'NYC'),
             (1, 'Alice', 100, 'Boston'),
             (4, 'Dennis', 400, 'Moscow')])

        query = compute(join(L, R, how='left'),
                        {L: left, R: right},
                        post_compute=False, return_type='native')
        result = list(map(tuple, conn.execute(query).fetchall()))

        assert set(result) == set(
            [(1, 'Alice', 100, 'NYC'),
             (1, 'Alice', 100, 'Boston'),
             (2, 'Bob', 200, None),
             (4, 'Dennis', 400, 'Moscow')])

        query = compute(join(L, R, how='right'),
                        {L: left, R: right},
                        post_compute=False, return_type='native')
        print(query)
        result = list(map(tuple, conn.execute(query).fetchall()))
        print(result)

        assert set(result) == set(
            [(1, 'Alice', 100, 'NYC'),
             (1, 'Alice', 100, 'Boston'),
             (3, None, None, 'LA'),
             (4, 'Dennis', 400, 'Moscow')])

        # SQLAlchemy doesn't support full outer join
        """
        query = compute(join(L, R, how='outer'),
                        {L: left, R: right},
                        post_compute=False)
        result = list(map(tuple, conn.execute(query).fetchall()))

        assert set(result) == set(
                [(1, 'Alice', 100, 'NYC'),
                 (1, 'Alice', 100, 'Boston'),
                 (2, 'Bob', 200, None),
                 (3, None, None, 'LA'),
                 (4, 'Dennis', 400, 'Moscow')])
        """

        conn.close()


def test_summary():
    expr = summary(a=t.amount.sum(), b=t.id.count())
    result = str(compute(expr, s, return_type='native'))

    assert 'sum(accounts.amount) as a' in result.lower()
    assert 'count(accounts.id) as b' in result.lower()


def test_summary_clean():
    t2 = t[t.amount > 0]
    expr = summary(a=t2.amount.sum(), b=t2.id.count())
    result = str(compute(expr, s, return_type='native'))

    assert normalize(result) == normalize("""
    SELECT sum(accounts.amount) as a, count(accounts.id) as b
    FROM accounts
    WHERE accounts.amount > :amount_1""")


def test_summary_by():
    expr = by(t.name, summary(a=t.amount.sum(), b=t.id.count()))

    result = str(compute(expr, s, return_type='native'))

    assert 'sum(accounts.amount) as a' in result.lower()
    assert 'count(accounts.id) as b' in result.lower()

    assert 'group by accounts.name' in result.lower()


def test_clean_join():
    metadata = sa.MetaData()
    name = sa.Table('name', metadata,
                    sa.Column('id', sa.Integer),
                    sa.Column('name', sa.String),
                    )
    city = sa.Table('place', metadata,
                    sa.Column('id', sa.Integer),
                    sa.Column('city', sa.String),
                    sa.Column('country', sa.String),
                    )
    friends = sa.Table('friends', metadata,
                       sa.Column('a', sa.Integer),
                       sa.Column('b', sa.Integer),
                       )

    tcity = symbol('city', discover(city))
    tfriends = symbol('friends', discover(friends))
    tname = symbol('name', discover(name))

    ns = {tname: name, tfriends: friends, tcity: city}

    expr = join(tfriends, tname, 'a', 'id')
    assert normalize(str(compute(expr, ns, return_type='native'))) == normalize("""
    SELECT friends.a, friends.b, name.name
    FROM friends JOIN name on friends.a = name.id""")

    expr = join(join(tfriends, tname, 'a', 'id'), tcity, 'a', 'id')

    result = compute(expr, ns, return_type='native')

    expected1 = """
    SELECT friends.a, friends.b, name.name, place.city, place.country
    FROM friends
        JOIN name ON friends.a = name.id
        JOIN place ON friends.a = place.id
        """

    expected2 = """
    SELECT alias.a, alias.b, alias.name, place.city, place.country
    FROM (SELECT friends.a AS a, friends.b AS b, name.name AS name
          FROM friends JOIN name ON friends.a = name.id) AS alias
    JOIN place ON alias.a = place.id
    """

    assert (normalize(str(result)) == normalize(expected1) or
            normalize(str(result)) == normalize(expected2))


def test_like():
    expr = t[t.name.like('Alice*')]
    assert normalize(str(compute(expr, s, return_type='native'))) == normalize("""
    SELECT accounts.name, accounts.amount, accounts.id
    FROM accounts
    WHERE accounts.name LIKE :name_1""")

def test_like_join():
    # Test .like() on a non-trivial SELECT query.
    metadata = sa.MetaData()
    t0 = sa.Table('t0', metadata,
                   sa.Column('a', sa.String),
                   sa.Column('y', sa.Integer))

    t1 = sa.Table('t1', metadata,
                  sa.Column('b', sa.String),
                  sa.Column('z', sa.Integer))

    # expected = t0.join(t1, t0.c.a == t1.c.b)
    # expected = sa.select(list(unique(expected.columns, key=lambda c:
                                  # c.name))).select_from(expected)
    # expected = expected.where(expected.c.a.like('%a'))

    t0sym = symbol('t0sym', 'var * {a: string, y: int}')
    t1sym = symbol('t1sym', 'var * {b: string, z: int}')
    joined = join(t0sym, t1sym, on_left='a', on_right='b')
    liked = joined[joined.a.like('*a')]
    result = compute(liked, {t0sym: t0, t1sym: t1}, return_type='native')
    assert normalize(str(result)) == normalize("""
    SELECT t0.a, t0.y, t1.z
    FROM t0 JOIN t1 ON t0.a = t1.b
    WHERE t0.a LIKE :a_1""")


def test_not_like():
    expr = t[~t.name.like('Alice*')]
    assert normalize(str(compute(expr, s, return_type='native'))) == normalize("""
    SELECT accounts.name, accounts.amount, accounts.id
    FROM accounts
    WHERE accounts.name NOT LIKE :name_1""")


def test_str_len():
    expr = t.name.str.len()
    result = str(compute(expr, s, return_type='native'))
    expected = "SELECT char_length(accounts.name) as name FROM accounts"
    assert normalize(result) == normalize(expected)


def test_str_upper():
    expr = t.name.str.upper()
    result = str(compute(expr, s, return_type='native'))
    expected = "SELECT upper(accounts.name) as name FROM accounts"
    assert normalize(result) == normalize(expected)


def test_str_lower():
    expr = t.name.str.lower()
    result = str(compute(expr, s, return_type='native'))
    expected = "SELECT lower(accounts.name) as name FROM accounts"
    assert normalize(result) == normalize(expected)


def test_str_find():
    expr = t.name.str.find('a')
    expected = """
               SELECT position(:param_1 in accounts.name) as position_1
               FROM accounts
               """
    result = str(compute(expr, s, return_type='native'))
    assert normalize(result) == normalize(expected)


def test_str_slice():
    expr = t.name.str[1:3]
    expected = """
               SELECT substring(accounts.name, :substring_2, :substring_3) as substring_1
               from accounts
               """
    result = str(compute(expr, s, return_type='native'))
    assert normalize(result) == normalize(expected)


@pytest.mark.parametrize("sep", [None, " sep "])
def test_str_cat(sep):
    """
    Need at least two string columns to test str_cat
    """

    if sep is None:
        expr = t_str_cat.name.str_cat(t_str_cat.comment)
        expected = """
                   SELECT accounts2.name || accounts2.comment
                   AS anon_1 FROM accounts2
                   """
    else:
        expr = t_str_cat.name.str_cat(t_str_cat.comment, sep=sep)
        expected = """
                   SELECT accounts2.name || :name_1 || accounts2.comment
                   AS anon_1 FROM accounts2
                   """

    result = str(compute(expr, s_str_cat, return_type='native'))
    assert normalize(result) == normalize(expected)


def test_str_cat_chain():
    expr = (t_str_cat.name
            .str_cat(t_str_cat.comment, sep=' -- ')
            .str_cat(t_str_cat.product, sep=' ++ '))
    result = str(compute(expr, {t_str_cat: s_str_cat}, return_type='native'))
    expected = """
               SELECT accounts2.name || :name_1 || accounts2.comment ||
               :param_1 || accounts2.product AS anon_1 FROM accounts2
               """
    assert normalize(result) == normalize(expected)


def test_str_cat_no_runtime_exception():
    """
    No exception raised if resource is the same
    """
    expr = t_str_cat.comment.str_cat(t.name)
    compute(expr, {t: s_str_cat, t_str_cat: s_str_cat}, return_type='native')


def test_columnwise_on_complex_selection():
    result = str(select(compute(t[t.amount > 0].amount + 1, s, return_type='native')))
    expected = """
               SELECT accounts.amount + :amount_1 AS amount
               FROM accounts
               WHERE accounts.amount > :amount_2
               """
    assert normalize(result) == normalize(expected)


def test_reductions_on_complex_selections():
    assert (
        normalize(str(select(compute(t[t.amount > 0].id.sum(), s, return_type='native')))) ==
        normalize(
            """
            select
                sum(alias.id) as id_sum
            from (select
                    accounts.id as id
                  from accounts
                  where accounts.amount > :amount_1) as alias
            """
        )
    )


def test_clean_summary_by_where():
    t2 = t[t.id == 1]
    expr = by(t2.name, sum=t2.amount.sum(), count=t2.amount.count())
    result = compute(expr, s, return_type='native')

    assert normalize(str(result)) == normalize("""
    SELECT accounts.name, count(accounts.amount) AS count, sum(accounts.amount) AS sum
    FROM accounts
    WHERE accounts.id = :id_1
    GROUP BY accounts.name
    """)


def test_by_on_count():
    expr = by(t.name, count=t.count())
    result = compute(expr, s, return_type='native')

    assert normalize(str(result)) == normalize("""
    SELECT accounts.name, count(accounts.id) AS count
    FROM accounts
    GROUP BY accounts.name
    """)


def test_join_complex_clean():
    metadata = sa.MetaData()
    name = sa.Table('name', metadata,
                    sa.Column('id', sa.Integer),
                    sa.Column('name', sa.String),
                    )
    city = sa.Table('place', metadata,
                    sa.Column('id', sa.Integer),
                    sa.Column('city', sa.String),
                    sa.Column('country', sa.String),
                    )

    tname = symbol('name', discover(name))
    tcity = symbol('city', discover(city))

    ns = {tname: name, tcity: city}

    expr = join(tname[tname.id > 0], tcity, 'id')
    result = compute(expr, ns, return_type='native')

    expected1 = """
        SELECT name.id, name.name, place.city, place.country
        FROM name JOIN place ON name.id = place.id
        WHERE name.id > :id_1"""

    expected2 = """
        SELECT alias.id, alias.name, place.city, place.country
        FROM (SELECT name.id as id, name.name AS name
              FROM name
              WHERE name.id > :id_1) AS alias
        JOIN place ON alias.id = place.id"""
    assert (normalize(str(result)) == normalize(expected1) or
            normalize(str(result)) == normalize(expected2))


def test_projection_of_join():
    metadata = sa.MetaData()
    name = sa.Table('name', metadata,
                    sa.Column('id', sa.Integer),
                    sa.Column('name', sa.String),
                    )
    city = sa.Table('place', metadata,
                    sa.Column('id', sa.Integer),
                    sa.Column('city', sa.String),
                    sa.Column('country', sa.String),
                    )

    tname = symbol('name', discover(name))
    tcity = symbol('city', discover(city))

    expr = join(tname, tcity[tcity.city == 'NYC'], 'id')[['country', 'name']]

    ns = {tname: name, tcity: city}

    result = compute(expr, ns, return_type='native')

    expected1 = """
        SELECT place.country, name.name
        FROM name JOIN place ON name.id = place.id
        WHERE place.city = :city_1"""

    expected2 = """
        SELECT alias.country, name.name
        FROM name
        JOIN (SELECT place.id AS id, place.city AS city, place.country AS country
              FROM place
              WHERE place.city = :city_1) AS alias
        ON name.id = alias_6.id"""

    assert (normalize(str(result)) == normalize(expected1) or
            normalize(str(result)) == normalize(expected2))


def test_lower_column():
    metadata = sa.MetaData()
    name = sa.Table('name', metadata,
                    sa.Column('id', sa.Integer),
                    sa.Column('name', sa.String),
                    )
    city = sa.Table('place', metadata,
                    sa.Column('id', sa.Integer),
                    sa.Column('city', sa.String),
                    sa.Column('country', sa.String),
                    )

    tname = symbol('name', discover(name))
    tcity = symbol('city', discover(city))

    assert lower_column(name.c.id) is name.c.id
    assert lower_column(select(name).c.id) is name.c.id

    j = name.join(city, name.c.id == city.c.id)
    col = [c for c in j.columns if c.name == 'country'][0]

    assert lower_column(col) is city.c.country


def test_selection_of_join():
    metadata = sa.MetaData()
    name = sa.Table('name', metadata,
                    sa.Column('id', sa.Integer),
                    sa.Column('name', sa.String),
                    )
    city = sa.Table('place', metadata,
                    sa.Column('id', sa.Integer),
                    sa.Column('city', sa.String),
                    sa.Column('country', sa.String),
                    )

    tname = symbol('name', discover(name))
    tcity = symbol('city', discover(city))

    ns = {tname: name, tcity: city}

    j = join(tname, tcity, 'id')
    expr = j[j.city == 'NYC'].name
    result = compute(expr, ns, return_type='native')

    assert normalize(str(result)) == normalize("""
    SELECT name.name
    FROM name JOIN place ON name.id = place.id
    WHERE place.city = :city_1""")


def test_join_on_same_table():
    metadata = sa.MetaData()
    T = sa.Table('tab', metadata,
                 sa.Column('a', sa.Integer),
                 sa.Column('b', sa.Integer),
                 )

    t = symbol('tab', discover(T))
    expr = join(t, t, 'a')

    result = compute(expr, {t: T}, return_type='native')

    assert normalize(str(result)) == normalize("""
    SELECT
        tab_left.a,
        tab_left.b as b_left,
        tab_right.b as b_right
    FROM
        tab AS tab_left
    JOIN
        tab AS tab_right
    ON
        tab_left.a = tab_right.a
    """)

    expr = join(t, t, 'a').b_left.sum()

    result = compute(expr, {t: T}, return_type='native')

    assert normalize(str(result)) == normalize("""
    select sum(alias.b_left) as b_left_sum from
        (select
            tab_left.b as b_left
        from
            tab as tab_left
        join
            tab as tab_right
        on
            tab_left.a = tab_right.a) as
        alias""")

    expr = join(t, t, 'a')
    expr = summary(total=expr.a.sum(), smallest=expr.b_right.min())

    result = compute(expr, {t: T}, return_type='native')

    assert normalize(str(result)) == normalize("""
    SELECT
        min(tab_right.b) as smallest,
        sum(tab_left.a) as total
    FROM
        tab AS tab_left
    JOIN
        tab AS tab_right
    ON
        tab_left.a = tab_right.a
    """)


def test_join_suffixes():
    metadata = sa.MetaData()
    T = sa.Table('tab', metadata,
                 sa.Column('a', sa.Integer),
                 sa.Column('b', sa.Integer),
                 )

    t = symbol('tab', discover(T))
    suffixes = '_l', '_r'
    expr = join(t, t, 'a', suffixes=suffixes)

    result = compute(expr, {t: T}, return_type='native')

    assert normalize(str(result)) == normalize("""
    SELECT
        tab{l}.a,
        tab{l}.b as b{l},
        tab{r}.b as b{r}
    FROM
        tab AS tab{l}
    JOIN
        tab AS tab{r}
    ON
        tab{l}.a = tab{r}.a
    """.format(l=suffixes[0], r=suffixes[1]))


def test_field_access_on_engines(city_data):
    s, engine = city_data['s'], city_data['engine']
    result = compute_up(s.city, engine, return_type='native')
    assert isinstance(result, sa.Table)
    assert result.name == 'city'


def test_computation_directly_on_sqlalchemy_Tables(city_data):
    name = city_data['name']
    s = symbol('s', discover(name))
    result = into(list, compute(s.id + 1, name, return_type='native'))
    assert not isinstance(result, sa.sql.Selectable)
    assert list(result) == []


def test_computation_directly_on_metadata(city_data):
    metadata = city_data['metadata']
    name = city_data['name']
    s = symbol('s', discover(metadata))
    result = compute(s.name, {s: metadata}, post_compute=False, return_type='native')
    assert result == name


sql_bank = sa.Table('bank', sa.MetaData(),
                    sa.Column('id', sa.Integer),
                    sa.Column('name', sa.String),
                    sa.Column('amount', sa.Integer))
sql_cities = sa.Table('cities', sa.MetaData(),
                      sa.Column('name', sa.String),
                      sa.Column('city', sa.String))

bank = symbol('bank', discover(sql_bank))
cities = symbol('cities', discover(sql_cities))


def test_aliased_views_with_two_group_bys():
    expr = by(bank.name, total=bank.amount.sum())
    expr2 = by(expr.total, count=expr.name.count())

    result = compute(expr2, {bank: sql_bank, cities: sql_cities}, return_type='native')

    assert normalize(str(result)) == normalize("""
    SELECT alias.total, count(alias.name) as count
    FROM (SELECT bank.name AS name, sum(bank.amount) AS total
          FROM bank
          GROUP BY bank.name) as alias
    GROUP BY alias.total
    """)


def test_aliased_views_with_join():
    joined = join(bank, cities)
    expr = by(joined.city, total=joined.amount.sum())
    expr2 = by(expr.total, count=expr.city.nunique())

    result = compute(expr2, {bank: sql_bank, cities: sql_cities}, return_type='native')

    assert normalize(str(result)) == normalize("""
    SELECT alias.total, count(DISTINCT alias.city) AS count
    FROM (SELECT cities.city AS city, sum(bank.amount) AS total
          FROM bank
          JOIN cities ON bank.name = cities.name
          GROUP BY cities.city) as alias
    GROUP BY alias.total
    """)


def test_select_field_on_alias():
    result = compute_up(t.amount, select(s).limit(10).alias('foo'))
    assert normalize(str(select(result))) == normalize("""
        SELECT foo.amount
        FROM (SELECT accounts.name AS name, accounts.amount AS amount, accounts.id AS id
              FROM accounts
              LIMIT :param_1) as foo""")


@pytest.mark.xfail(raises=Exception,
                   reason="sqlalchemy.join seems to drop unnecessary tables")
def test_join_on_single_column():
    expr = join(cities[['name']], bank)
    result = compute(expr, {bank: sql_bank, cities: sql_cities}, return_type='native')

    assert normalize(str(result)) == """
    SELECT bank.id, bank.name, bank.amount
    FROM bank join cities ON bank.name = cities.name"""

    expr = join(bank, cities.name)
    result = compute(expr, {bank: sql_bank, cities: sql_cities}, return_type='native')

    assert normalize(str(result)) == """
    SELECT bank.id, bank.name, bank.amount
    FROM bank join cities ON bank.name = cities.name"""


def test_aliased_views_more():
    metadata = sa.MetaData()
    lhs = sa.Table('aaa', metadata,
                   sa.Column('x', sa.Integer),
                   sa.Column('y', sa.Integer),
                   sa.Column('z', sa.Integer))

    rhs = sa.Table('bbb', metadata,
                   sa.Column('w', sa.Integer),
                   sa.Column('x', sa.Integer),
                   sa.Column('y', sa.Integer))

    L = symbol('L', 'var * {x: int, y: int, z: int}')
    R = symbol('R', 'var * {w: int, x: int, y: int}')

    expr = join(by(L.x, y_total=L.y.sum()),
                R)

    result = compute(expr, {L: lhs, R: rhs}, return_type='native')

    assert normalize(str(result)) == normalize("""
        SELECT alias.x, alias.y_total, bbb.w, bbb.y
        FROM (SELECT aaa.x as x, sum(aaa.y) as y_total
              FROM aaa
              GROUP BY aaa.x) AS alias
        JOIN bbb ON alias.x = bbb.x """)

    expr2 = by(expr.w, count=expr.x.count(), total2=expr.y_total.sum())

    result2 = compute(expr2, {L: lhs, R: rhs}, return_type='native')

    assert (
        normalize(str(result2)) == normalize("""
            SELECT alias_2.w, count(alias_2.x) as count, sum(alias_2.y_total) as total2
            FROM (SELECT alias.x, alias.y_total, bbb.w, bbb.y
                  FROM (SELECT aaa.x as x, sum(aaa.y) as y_total
                        FROM aaa
                        GROUP BY aaa.x) AS alias
                  JOIN bbb ON alias.x = bbb.x) AS alias_2
            GROUP BY alias_2.w""")

        or

        normalize(str(result2)) == normalize("""
            SELECT bbb.w, count(alias.x) as count, sum(alias.y_total) as total2
            FROM (SELECT aaa.x as x, sum(aaa.y) as y_total
                  FROM aaa
                  GROUP BY aaa.x) as alias
              JOIN bbb ON alias.x = bbb.x
            GROUP BY bbb.w"""))


def test_aliased_views_with_computation():
    engine = sa.create_engine('sqlite:///:memory:')

    df_aaa = DataFrame({'x': [1, 2, 3, 2, 3],
                        'y': [2, 1, 2, 3, 1],
                        'z': [3, 3, 3, 1, 2]})
    df_bbb = DataFrame({'w': [1, 2, 3, 2, 3],
                        'x': [2, 1, 2, 3, 1],
                        'y': [3, 3, 3, 1, 2]})

    df_aaa.to_sql('aaa', engine)
    df_bbb.to_sql('bbb', engine)

    metadata = sa.MetaData(engine)
    metadata.reflect()

    sql_aaa = metadata.tables['aaa']
    sql_bbb = metadata.tables['bbb']

    L = symbol('aaa', discover(df_aaa))
    R = symbol('bbb', discover(df_bbb))

    expr = join(by(L.x, y_total=L.y.sum()),
                R)
    a = compute(expr, {L: df_aaa, R: df_bbb}, return_type='native')
    b = compute(expr, {L: sql_aaa, R: sql_bbb}, return_type='native')
    assert into(set, a) == into(set, b)

    expr2 = by(expr.w, count=expr.x.count(), total2=expr.y_total.sum())
    a = compute(expr2, {L: df_aaa, R: df_bbb}, return_type='native')
    b = compute(expr2, {L: sql_aaa, R: sql_bbb}, return_type='native')
    assert into(set, a) == into(set, b)

    expr3 = by(expr.x, count=expr.y_total.count())
    a = compute(expr3, {L: df_aaa, R: df_bbb}, return_type='native')
    b = compute(expr3, {L: sql_aaa, R: sql_bbb}, return_type='native')
    assert into(set, a) == into(set, b)

    expr4 = join(expr2, R)
    a = compute(expr4, {L: df_aaa, R: df_bbb}, return_type='native')
    b = compute(expr4, {L: sql_aaa, R: sql_bbb}, return_type='native')
    assert into(set, a) == into(set, b)

    """ # Takes a while
    expr5 = by(expr4.count, total=(expr4.x + expr4.y).sum())
    a = compute(expr5, {L: df_aaa, R: df_bbb})
    b = compute(expr5, {L: sql_aaa, R: sql_bbb})
    assert into(set, a) == into(set, b)
    """


def test_distinct_count_on_projection():
    expr = t[['amount']].distinct().count()

    result = compute(expr, {t: s}, return_type='native')

    assert (
        normalize(str(result)) == normalize("""
        SELECT count(DISTINCT accounts.amount)
        FROM accounts""")

        or

        normalize(str(result)) == normalize("""
        SELECT count(alias.amount) as count
        FROM (SELECT DISTINCT accounts.amount AS amount
              FROM accounts) as alias"""))

    # note that id is the primary key
    expr = t[['amount', 'id']].distinct().count()

    result = compute(expr, {t: s}, return_type='native')
    assert normalize(str(result)) == normalize("""
        SELECT count(alias.id) as count
        FROM (SELECT DISTINCT accounts.amount AS amount, accounts.id AS id
              FROM accounts) as alias""")


def test_join_count():
    ds = datashape.dshape(
        '{t1: var * {x: int, y: int}, t2: var * {a: int, b: int}}')
    engine = bz_data('sqlite:///:memory:', dshape=ds)
    db = symbol('db', ds)

    expr = join(db.t1[db.t1.x > -1], db.t2, 'x', 'a').count()

    result = compute(expr, {db: engine}, post_compute=False, return_type='native')

    expected1 = """
    SELECT count(alias.x) as count
    FROM (SELECT t1.x AS x, t1.y AS y, t2.b AS b
          FROM t1 JOIN t2 ON t1.x = t2.a
          WHERE t1.x > ?) as alias
          """
    expected2 = """
    SELECT count(alias2.x) AS count
    FROM (SELECT alias1.x AS x, alias1.y AS y, t2.b AS b
          FROM (SELECT t1.x AS x, t1.y AS y
                FROM t1
                WHERE t1.x > ?) AS alias1
          JOIN t2 ON alias1.x = t2.a) AS alias2"""

    assert (normalize(str(result)) == normalize(expected1) or
            normalize(str(result)) == normalize(expected2))


def test_transform_where():
    t2 = t[t.id == 1]
    expr = transform(t2, abs_amt=abs(t2.amount), sine=sin(t2.id))
    result = compute(expr, s, return_type='native')

    expected = """SELECT
        accounts.name,
        accounts.amount,
        accounts.id,
        abs(accounts.amount) as abs_amt,
        sin(accounts.id) as sine
    FROM accounts
    WHERE accounts.id = :id_1
    """

    assert normalize(str(result)) == normalize(expected)


def test_merge():
    col = (t['amount'] * 2).label('new')

    expr = merge(t['name'], col)

    result = str(compute(expr, s, return_type='native'))

    assert 'amount * ' in result
    assert 'FROM accounts' in result
    assert 'SELECT accounts.name' in result
    assert 'new' in result


def test_merge_where():
    t2 = t[t.id == 1]
    expr = merge(t2[['amount', 'name']], t2.id)
    result = compute(expr, s, return_type='native')
    expected = normalize("""SELECT
        accounts.amount,
        accounts.name,
        accounts.id
    FROM accounts
    WHERE accounts.id = :id_1
    """)
    assert normalize(str(result)) == expected


def test_transform_filter_by_single_column():
    t2 = t[t.amount < 0]
    tr = transform(t2, abs_amt=abs(t2.amount), sine=sin(t2.id))
    expr = by(tr.name, avg_amt=tr.abs_amt.mean())
    result = compute(expr, s, return_type='native')
    expected = normalize("""SELECT
        accounts.name,
        avg(abs(accounts.amount)) AS avg_amt
    FROM accounts
    WHERE accounts.amount < :amount_1
    GROUP BY accounts.name
    """)
    assert normalize(str(result)) == expected


def test_transform_filter_by_multiple_columns():
    t2 = t[t.amount < 0]
    tr = transform(t2, abs_amt=abs(t2.amount), sine=sin(t2.id))
    expr = by(tr.name, avg_amt=tr.abs_amt.mean(), sum_sine=tr.sine.sum())
    result = compute(expr, s, return_type='native')
    expected = normalize("""SELECT
        accounts.name,
        avg(abs(accounts.amount)) AS avg_amt,
        sum(sin(accounts.id)) AS sum_sine
    FROM accounts
    WHERE accounts.amount < :amount_1
    GROUP BY accounts.name
    """)
    assert normalize(str(result)) == expected


def test_transform_filter_by_different_order():
    t2 = transform(t, abs_amt=abs(t.amount), sine=sin(t.id))
    tr = t2[t2.amount < 0]
    expr = by(tr.name,
              avg_amt=tr.abs_amt.mean(),
              avg_sine=tr.sine.sum() / tr.sine.count())
    result = compute(expr, s, return_type='native')
    expected = normalize("""SELECT
        accounts.name,
        avg(abs(accounts.amount)) AS avg_amt,
        sum(sin(accounts.id)) / count(sin(accounts.id)) AS avg_sine
    FROM accounts
    WHERE accounts.amount < :amount_1
    GROUP BY accounts.name
    """)
    assert normalize(str(result)) == expected


def test_transform_filter_by_projection():
    t2 = transform(t, abs_amt=abs(t.amount), sine=sin(t.id))
    tr = t2[t2.amount < 0]
    expr = by(tr[['name', 'id']],
              avg_amt=tr.abs_amt.mean(),
              avg_sine=tr.sine.sum() / tr.sine.count())
    result = compute(expr, s, return_type='native')
    expected = normalize("""SELECT
        accounts.name,
        accounts.id,
        avg(abs(accounts.amount)) AS avg_amt,
        sum(sin(accounts.id)) / count(sin(accounts.id)) AS avg_sine
    FROM accounts
    WHERE accounts.amount < :amount_1
    GROUP BY accounts.name, accounts.id
    """)
    assert normalize(str(result)) == expected


def test_merge_compute():
    dta = [(1, 'Alice', 100),
            (2, 'Bob', 200),
            (4, 'Dennis', 400)]
    ds = datashape.dshape('var * {id: int, name: string, amount: real}')
    s = symbol('s', ds)

    with tmpfile('db') as fn:
        uri = 'sqlite:///' + fn
        into(uri + '::table', dta, dshape=ds)

        expr = transform(s, amount10=s.amount * 10)
        result = into(list, compute(expr, {s: dta}, return_type='native'))

        assert result == [(1, 'Alice', 100, 1000),
                          (2, 'Bob', 200, 2000),
                          (4, 'Dennis', 400, 4000)]


def test_notnull():
    result = compute(nt[nt.name.notnull()], ns, return_type='native')
    expected = """SELECT
        nullaccounts.name,
        nullaccounts.amount,
        nullaccounts.id
    FROM nullaccounts
    WHERE nullaccounts.name is not null
    """
    assert normalize(str(result)) == normalize(expected)


def test_head_limit():
    assert compute(t.head(5).head(10), s, return_type='native')._limit == 5
    assert compute(t.head(10).head(5), s, return_type='native')._limit == 5
    assert compute(t.head(10).head(10), s, return_type='native')._limit == 10


def test_no_extraneous_join():
    ds = """ {event: var * {name: ?string,
                            operation: ?string,
                            datetime_nearest_receiver: ?datetime,
                            aircraft: ?string,
                            temperature_2m: ?float64,
                            temperature_5cm: ?float64,
                            humidity: ?float64,
                            windspeed: ?float64,
                            pressure: ?float64,
                            include: int64},
             operation: var * {name: ?string,
                               runway: int64,
                               takeoff: bool,
                               datetime_nearest_close: ?string}}
        """
    db = bz_data('sqlite:///:memory:', dshape=ds)

    d = symbol('db', dshape=ds)

    expr = join(d.event[d.event.include == True],
                d.operation[['name', 'datetime_nearest_close']],
                'operation', 'name')

    result = compute(expr, db, return_type='native')

    assert normalize(str(result)) == normalize("""
    SELECT
        alias.operation,
        alias.name as name_left,
        alias.datetime_nearest_receiver,
        alias.aircraft,
        alias.temperature_2m,
        alias.temperature_5cm,
        alias.humidity,
        alias.windspeed,
        alias.pressure,
        alias.include,
        alias.datetime_nearest_close
    FROM
        (SELECT
             event.name AS name,
             event.operation AS operation,
             event.datetime_nearest_receiver AS datetime_nearest_receiver,
             event.aircraft AS aircraft,
             event.temperature_2m AS temperature_2m,
             event.temperature_5cm AS temperature_5cm,
             event.humidity AS humidity,
             event.windspeed AS windspeed,
             event.pressure AS pressure,
             event.include AS include
         FROM
             event WHERE event.include = 1) AS alias1
    JOIN
        (SELECT
             operation.name AS name,
             operation.datetime_nearest_close as datetime_nearest_close
         FROM operation) AS alias2
    ON
        alias1.operation = alias2.name
    """)


def test_math():
    result = compute(sin(t.amount), s, return_type='native')
    assert normalize(str(result)) == normalize("""
            SELECT sin(accounts.amount) as amount
            FROM accounts""")

    result = compute(floor(t.amount), s, return_type='native')
    assert normalize(str(result)) == normalize("""
            SELECT floor(accounts.amount) as amount
            FROM accounts""")

    result = compute(t.amount // 2, s, return_type='native')
    assert normalize(str(result)) == normalize("""
            SELECT floor(accounts.amount / :amount_1) AS amount
            FROM accounts""")


def test_transform_order():
    r = transform(t, sin_amount=sin(t.amount), cos_id=cos(t.id))
    result = compute(r, s, return_type='native')
    expected = """SELECT
        accounts.name,
        accounts.amount,
        accounts.id,
        cos(accounts.id) as cos_id,
        sin(accounts.amount) as sin_amount
    FROM accounts
    """
    assert normalize(str(result)) == normalize(expected)


def test_isin():
    result = t[t.name.isin(['foo', 'bar'])]
    result_sql_expr = str(compute(result, s, return_type='native'))
    expected = """
        SELECT
            accounts.name,
            accounts.amount,
            accounts.id
        FROM
            accounts
        WHERE
            accounts.name
        IN
            (:name_1,
            :name_2)
    """
    assert normalize(result_sql_expr) == normalize(expected)


@pytest.mark.skipif('1.0.0' <= LooseVersion(sa.__version__) <= '1.0.1',
                    reason=("SQLAlchemy generates different code in 1.0.0"
                            " and 1.0.1"))
def test_date_grouper_repeats_not_one_point_oh():
    columns = [sa.Column('amount', sa.REAL),
               sa.Column('ds', sa.TIMESTAMP)]
    dta = sa.Table('t', sa.MetaData(), *columns)
    t = symbol('t', discover(dta))
    expr = by(t.ds.year, avg_amt=t.amount.mean())
    result = str(compute(expr, dta, return_type='native'))

    # FYI spark sql isn't able to parse this correctly
    expected = """SELECT
        EXTRACT(year FROM t.ds) as ds_year,
        AVG(t.amount) as avg_amt
    FROM t
    GROUP BY EXTRACT(year FROM t.ds)
    """
    assert normalize(result) == normalize(expected)


@pytest.mark.skipif(LooseVersion(sa.__version__) < '1.0.0' or
                    LooseVersion(sa.__version__) >= '1.0.2',
                    reason=("SQLAlchemy generates different code in < 1.0.0 "
                            "and >= 1.0.2"))
def test_date_grouper_repeats():
    columns = [sa.Column('amount', sa.REAL),
               sa.Column('ds', sa.TIMESTAMP)]
    dta = sa.Table('t', sa.MetaData(), *columns)
    t = symbol('t', discover(dta))
    expr = by(t.ds.year, avg_amt=t.amount.mean())
    result = str(compute(expr, dta, return_type='native'))

    # FYI spark sql isn't able to parse this correctly
    expected = """SELECT
        EXTRACT(year FROM t.ds) as ds_year,
        AVG(t.amount) as avg_amt
    FROM t
    GROUP BY ds_year
    """
    assert normalize(result) == normalize(expected)


def test_transform_then_project_single_column():
    expr = transform(t, foo=t.id + 1)[['foo', 'id']]
    result = normalize(str(compute(expr, s, return_type='native')))
    expected = normalize("""SELECT
        accounts.id + :id_1 as foo,
        accounts.id
    FROM accounts""")
    assert result == expected


def test_transform_then_project():
    proj = ['foo', 'id']
    expr = transform(t, foo=t.id + 1)[proj]
    result = normalize(str(compute(expr, s, return_type='native')))
    expected = normalize("""SELECT
        accounts.id + :id_1 as foo,
        accounts.id
    FROM accounts""")
    assert result == expected


def test_reduce_does_not_compose():
    expr = by(t.name, counts=t.count()).counts.max()
    result = str(compute(expr, s, return_type='native'))
    expected = """
    SELECT max(alias.counts) AS counts_max
    FROM
    (SELECT count(accounts.id) AS counts
    FROM accounts GROUP BY accounts.name) as alias"""
    assert normalize(result) == normalize(expected)


@pytest.mark.xfail(raises=NotImplementedError)
def test_normalize_reduction():
    expr = by(t.name, counts=t.count())
    expr = transform(expr, normed_counts=expr.counts / expr.counts.max())
    result = str(compute(expr, s, return_type='native'))
    expected = """WITH alias AS
(SELECT count(accounts.id) AS counts
FROM accounts GROUP BY accounts.name)
 SELECT alias.counts / max(alias.counts) AS normed_counts
FROM alias"""
    assert normalize(result) == normalize(expected)


def test_do_not_erase_group_by_functions_with_datetime():
    t, s = tdate, sdate
    expr = by(t[t.amount < 0].occurred_on.date,
              avg_amount=t[t.amount < 0].amount.mean())
    result = str(compute(expr, s, return_type='native'))
    expected = """SELECT
        date(accdate.occurred_on) as occurred_on_date,
        avg(accdate.amount) as avg_amount
    FROM
        accdate
    WHERE
        accdate.amount < :amount_1
    GROUP BY
        date(accdate.occurred_on)
    """
    assert normalize(result) == normalize(expected)


def test_not():
    expr = t.amount[~t.name.isin(('Billy', 'Bob'))]
    result = str(compute(expr, s, return_type='native'))
    expected = """SELECT
        accounts.amount
    FROM
        accounts
    WHERE
        accounts.name not in (:name_1, :name_2)
    """
    assert normalize(result) == normalize(expected)


def test_slice():
    start, stop, step = 50, 100, 1
    result = str(compute(t[start:stop], s, return_type='native'))

    # Verifies that compute is translating the query correctly
    assert result == str(select(s).offset(start).limit(stop))

    # Verifies the query against expected SQL query
    expected = """
    SELECT accounts.name, accounts.amount, accounts.id FROM accounts
    LIMIT :param_1 OFFSET :param_2
    """

    assert normalize(str(result)) == normalize(str(expected))

    # Step size of 1 should be alright
    compute(t[start:stop:step], s, return_type='native')


@pytest.mark.xfail(raises=ValueError)
def test_slice_step():
    start, stop, step = 50, 100, 2
    compute(t[start:stop:step], s, return_type='native')


def test_datetime_to_date():
    expr = tdate.occurred_on.date
    result = str(compute(expr, sdate, return_type='native'))
    expected = """SELECT
        DATE(accdate.occurred_on) as occurred_on_date
    FROM
        accdate
    """
    assert normalize(result) == normalize(expected)


def test_sort_compose():
    expr = t.name[:5].sort()
    result = compute(expr, s, return_type='native')
    expected = """select
            anon_1.name
        from (select
                  accounts.name as name
              from
                  accounts
              limit :param_1
              offset :param_2) as anon_1
        order by
            anon_1.name asc"""
    assert normalize(str(result)) == normalize(expected)
    assert (normalize(str(compute(t.sort('name').name[:5], s, return_type='native'))) !=
            normalize(expected))


def test_coerce():
    expr = t.amount.coerce(to='int64')
    expected = """SELECT
        cast(accounts.amount AS BIGINT) AS amount
    FROM accounts"""
    result = compute(expr, s, return_type='native')
    assert normalize(str(result)) == normalize(expected)


def test_multi_column_by_after_transform():
    tbl = transform(t, new_amount=t.amount + 1, one_two=t.amount * 2)
    expr = by(tbl[['name', 'one_two']], avg_amt=tbl.new_amount.mean())
    result = compute(expr, s, return_type='native')
    expected = """SELECT
        accounts.name,
        accounts.amount * :amount_1 as one_two,
        avg(accounts.amount + :amount_2) as avg_amt
    FROM
        accounts
    GROUP BY
        accounts.name, accounts.amount * :amount_1
    """
    assert normalize(str(result)) == normalize(expected)


def test_multi_column_by_after_transform_and_filter():
    tbl = t[t.name == 'Alice']
    tbl = transform(tbl, new_amount=tbl.amount + 1, one_two=tbl.amount * 2)
    expr = by(tbl[['name', 'one_two']], avg_amt=tbl.new_amount.mean())
    result = compute(expr, s, return_type='native')
    expected = """SELECT
        accounts.name,
        accounts.amount * :amount_1 as one_two,
        avg(accounts.amount + :amount_2) as avg_amt
    FROM
        accounts
    WHERE
        accounts.name = :name_1
    GROUP BY
        accounts.name, accounts.amount * :amount_1
    """
    assert normalize(str(result)) == normalize(expected)


def test_attribute_access_on_transform_filter():
    tbl = transform(t, new_amount=t.amount + 1)
    expr = tbl[tbl.name == 'Alice'].new_amount
    result = compute(expr, s, return_type='native')
    expected = """SELECT
        accounts.amount + :amount_1 as new_amount
    FROM
        accounts
    WHERE
        accounts.name = :name_1
    """
    assert normalize(str(result)) == normalize(expected)


def test_attribute_on_filter_transform_groupby():
    tbl = t[t.name == 'Alice']
    tbl = transform(tbl, new_amount=tbl.amount + 1, one_two=tbl.amount * 2)
    gb = by(tbl[['name', 'one_two']], avg_amt=tbl.new_amount.mean())
    expr = gb.avg_amt
    result = compute(expr, s, return_type='native')
    expected = """SELECT
        avg(accounts.amount + :amount_1) as avg_amt
    FROM
        accounts
    WHERE
        accounts.name = :name_1
    GROUP BY
        accounts.name, accounts.amount * :amount_2
    """
    assert normalize(str(result)) == normalize(expected)


def test_label_projection():
    tbl = t[(t.name == 'Alice')]
    tbl = transform(tbl, new_amount=tbl.amount + 1, one_two=tbl.amount * 2)
    expr = tbl[['new_amount', 'one_two']]

    # column selection shouldn't affect the resulting SQL
    result = compute(expr[expr.new_amount > 1].one_two, s, return_type='native')
    result2 = compute(expr.one_two[expr.new_amount > 1], s, return_type='native')

    expected = """SELECT
        accounts.amount * :amount_1 as one_two
    FROM accounts
    WHERE accounts.name = :name_1 and accounts.amount + :amount_2 > :param_1
    """
    assert normalize(str(result)) == normalize(expected)
    assert normalize(str(result2)) == normalize(expected)


def test_baseball_nested_by():
    dta = bz_data('sqlite:///%s' % example('teams.db'))
    dshape = discover(dta)
    d = symbol('d', dshape)
    expr = by(d.teams.name,
              start_year=d.teams.yearID.min()).start_year.count_values()
    result = compute(expr, dta, post_compute=False, return_type='native')
    expected = """SELECT
        anon_1.start_year,
        anon_1.count
    FROM
        (SELECT
            alias.start_year as start_year,
            count(alias.start_year) as count
         FROM
            (SELECT
                min(teams.yearid) as start_year
             FROM teams
             GROUP BY teams.name) as alias
         GROUP BY alias.start_year) as anon_1 ORDER BY anon_1.count DESC
    """
    assert normalize(str(result).replace('"', '')) == normalize(expected)


def test_label_on_filter():
    expr = t[t.name == 'Alice'].amount.label('foo').head(2)
    result = compute(expr, s, return_type='native')
    expected = """SELECT
        accounts.amount AS foo
    FROM
        accounts
    WHERE
        accounts.name = :name_1
    LIMIT :param_1
    """
    assert normalize(str(result)) == normalize(expected)


def test_single_field_filter():
    expr = t.amount[t.amount > 0]
    result = compute(expr, s, return_type='native')
    expected = """SELECT
        accounts.amount
    FROM accounts
    WHERE accounts.amount > :amount_1
    """
    assert normalize(str(result)) == normalize(expected)


def test_multiple_field_filter():
    expr = t.name[t.amount > 0]
    result = compute(expr, s, return_type='native')
    expected = """SELECT
        accounts.name
    FROM accounts
    WHERE accounts.amount > :amount_1
    """
    assert normalize(str(result)) == normalize(expected)


def test_distinct_on_label():
    expr = t.name.label('foo').distinct()
    result = compute(expr, s, return_type='native')
    expected = """SELECT
        DISTINCT accounts.name AS foo
    FROM accounts
    """
    assert normalize(str(result)) == normalize(expected)


@pytest.mark.parametrize('n', [-1, 0, 1])
def test_shift_on_column(n):
    expr = t.name.shift(n)
    result = compute(expr, s, return_type='native')
    expected = """SELECT
        lag(accounts.name, :lag_1) over () as name
    FROM accounts
    """
    assert normalize(str(result)) == normalize(expected)


@pytest.fixture
def lhs():
    return pd.DataFrame.from_records(
        dict(zip('cd', row)) for row in zip(
            ['AAPL', 'FB', 'AMZN', 'FB', 'GOOG'] * 2,
            list(np.random.rand(10) * 100)
        )
    )


@pytest.yield_fixture
def rhs():
    try:
        t = bz_data(
            'sqlite:///:memory:::rhs',
            dshape='var * {a: int64, b: ?float64, c: ?string}'
        )
    except sa.exc.OperationalError as e:
        pytest.skip(str(e))
    else:
        try:
            t.insert().values([
                dict(zip(t.columns.keys(), row)) for row in zip(
                    range(5),
                    list(np.random.rand(5) * 100),
                    ['AAPL', 'AAPL', 'MSFT', 'FB', 'MSFT']
                )
            ]).execute()
            yield t
        finally:
            drop(t)



@pytest.mark.skipif(LooseVersion(sqlite3.version) <= '2.6.0',
                    reason="Older sqlite3 versions don't support this")
def test_multiple_table_join(lhs, rhs):
    lexpr = symbol('lhs', discover(lhs))
    rexpr = symbol('rhs', discover(rhs))
    expr = join(lexpr, rexpr, 'c')
    result = compute(expr, {lexpr: lhs, rexpr: rhs}, return_type='native')
    expected = pd.merge(lhs, odo(rhs, pd.DataFrame), on='c')
    tm.assert_frame_equal(result, expected)


def test_empty_string_comparison_with_option_type():
    expr = nt.amount[nt.name == '']
    result = compute(expr, s, return_type='native')
    expected = """
    SELECT accounts.amount
    FROM accounts
    WHERE accounts.name = :name_1
    """
    assert normalize(str(result)) == normalize(expected)


def test_tail_no_sort():
    assert (
        normalize(str(compute(t.head(), {t: s}, return_type='native'))) ==
        normalize(str(compute(t.tail(), {t: s}, return_type='native')))
    )


def test_tail_of_sort():
    expected = normalize(str(compute(
        t.sort('id', ascending=False).head(5).sort('id'),
        {t: s}, return_type='native'
    )))
    result = normalize(str(compute(t.sort('id').tail(5), {t: s}, return_type='native')))
    assert expected == result


def test_tail_sort_in_chilren():
    expected = normalize(str(compute(
        t.name.sort('id', ascending=False).head(5).sort('id'),
        {t: s}, return_type='native'
    )))
    result = normalize(str(compute(t.name.sort('id').tail(5), {t: s}, return_type='native')))
    assert expected == result


def test_selection_inner_inputs():
    result = normalize(str(compute(t[t.id == tdate.id], {t: s, tdate: sdate}, return_type='native')))
    expected = normalize("""
    select {a}.name, {a}.amount, {a}.id from {a}, {b} where {a}.id = {b}.id
    """).format(a=s.name, b=sdate.name)
    assert result == expected


@pytest.mark.parametrize('unit', bz_datetime.units)
def test_datetime_trunc(unit):
    ds = 'var * {a: datetime}'
    db = bz_data('sqlite:///:memory:::s', dshape=ds)
    s = symbol('s', ds)

    assert normalize(compute(s.a.truncate(**{unit: 1}), db, return_type='native')) == normalize(
        'select date_trunc(%r, s.a) as a from s' % unit,
    )


@pytest.mark.parametrize('alias,unit', bz_datetime.unit_aliases.items())
def test_datetime_trunc_aliases(alias, unit):
    ds = 'var * {a: datetime}'
    db = bz_data('sqlite:///:memory:::s', dshape=ds)
    s = symbol('s', ds)

    assert normalize(compute(s.a.truncate(**{alias: 1}), db, return_type='native')) == normalize(
        'select date_trunc(%r, s.a) as a from s' % unit,
    )


@pytest.mark.xfail(
    raises=AssertionError,
    reason="We don't currently handle the inner select",
)
def test_inner_select_with_filter():
    ds = 'var * {a: float32}'
    db = bz_data('sqlite:///:memory:::s', dshape=ds)
    s = symbol('s', ds)

    assert normalize(compute(s[s.a == s.a[s.a == 1]].a, db, return_type='native')) == normalize(
        """\
        select
            s.a
        from s, (
            select
                s.a as a
            from
                s
            where
                s.a = 1
        ) as anon_1
        where
            s.a = anon_1.a
        """,
    )


@pytest.mark.parametrize('op', (greatest, least))
def test_greatest(op):
    ds = 'var * {a: int32, b: int32}'
    db = bz_data('sqlite:///:memory:::s', dshape=ds)
    odo([(1, 2)], db)
    s = symbol('s', dshape=ds)

    assert normalize(compute(op(s.a, s.b), db)) == normalize(
        'select {op}(s.a, s.b) as {op}_1 from s'.format(op=op.__name__),
    )


def test_coalesce():
    ds = 'var * {a: ?int32}'
    db = resource('sqlite:///:memory:::s', dshape=ds)
    s = symbol('s', ds)
    t = symbol('t', 'int32')

    assert normalize(compute(coalesce(s.a, t), {s: db, t: 1})) == normalize(
        """\
        select
            coalesce(s.a, 1) as coalesce_1
        from s
        """,
    )
    assert normalize(compute(coalesce(s.a, 1), {s: db})) == normalize(
        """
        select
            coalesce(s.a, 1) as a
        from s
        """,
    )


def test_any():
    ds = datashape.dshape('var * {a: int32}')
    db = resource('sqlite:///:memory:::s', dshape=ds)
    s = symbol('s', ds)
    t = symbol('t', 'int32')

    assert normalize(compute((s.a == t).any(), {s: db, t: 1})) == normalize(
        """
        select
            sum(cast(s.a = 1 as integer)) != 0 as anon_1
        from s
        """,
    )


def test_all():
    ds = datashape.dshape('var * {a: int32}')
    db = resource('sqlite:///:memory:::s', dshape=ds)
    s = symbol('s', ds)
    t = symbol('t', 'int32')

    assert normalize(compute((s.a == t).all(), {s: db, t: 1})) == normalize(
        """
        select
            sum(cast(s.a != 1 as integer)) = 0 as anon_1
        from s
        """,
    )



from __future__ import absolute_import, division, print_function

import warnings

from datashape import dshape
import flask
from flask.testing import FlaskClient
from odo import resource
import requests
from requests.packages.urllib3.exceptions import InsecureRequestWarning
from toolz import assoc, keymap

from ..compatibility import u8
from ..expr import Expr
from ..dispatch import dispatch
from .server import DEFAULT_PORT
from .serialization import json

# These are a hack for testing
# It's convenient to use requests for live production but use
# flask for testing.  Sadly they have different Response objects,
# hence the dispatched functions

__all__ = 'Client',


def content(response):
    if isinstance(response, flask.Response):
        return response.data
    if isinstance(response, requests.Response):
        return response.content


def ok(response):
    if isinstance(response, flask.Response):
        return 200 <= response.status_code <= 299
    if isinstance(response, requests.Response):
        return response.ok


def reason(response):
    if isinstance(response, flask.Response):
        return response.status
    if isinstance(response, requests.Response):
        return response.text


def _request(method, client, url, params=None, auth=None, **kwargs):
    if not isinstance(requests, FlaskClient):
        kwargs['verify'] = client.verify_ssl
        kwargs['params'] = params
        kwargs['auth'] = auth

    with warnings.catch_warnings():
        warnings.simplefilter('ignore', InsecureRequestWarning)
        return method(
            '{base}{url}'.format(
                base=client.url,
                url=url,
            ),
            **kwargs
        )


def get(*args, **kwargs):
    return _request(requests.get, *args, **kwargs)


def post(*args, **kwargs):
    return _request(requests.post, *args, **kwargs)


class Client(object):

    """ Client for Blaze Server

    Provides programmatic access to datasets living on Blaze Server

    Parameters
    ----------

    url : str
        URL of a Blaze server
    serial : SerializationFormat, optional
        The serialization format object to use. Defaults to JSON.
        A serialization format is an object that supports:
        name, loads, and dumps.
    verify_ssl : bool, optional
        Verify the ssl certificate from the server.
        This is enabled by default.
    auth : tuple, optional
        The username and password to use when connecting to the server.
        If not provided, no auth header will be sent.

    Examples
    --------

    >>> # This example matches with the docstring of ``Server``
    >>> from blaze import data
    >>> c = Client('localhost:6363')
    >>> t = data(c) # doctest: +SKIP

    See Also
    --------

    blaze.server.server.Server
    """
    __slots__ = 'url', 'serial', 'verify_ssl', 'auth'

    def __init__(self, url, serial=json, verify_ssl=True, auth=None, **kwargs):
        url = url.strip('/')
        if not url.startswith('http'):
            url = 'http://' + url
        self.url = url
        self.serial = serial
        self.verify_ssl = verify_ssl
        self.auth = auth

    @property
    def dshape(self):
        """The datashape of the client"""
        response = get(self, '/datashape', auth=self.auth)
        if not ok(response):
            raise ValueError("Bad Response: %s" % reason(response))

        return dshape(content(response).decode('utf-8'))

    def add(self, name, resource_uri, *args, **kwargs):
        """Add the given resource URI to the Blaze server.

        Parameters
        ----------
        name : str
            The name to give the resource
        resource_uri : str
            The URI string describing the resource to add to the server, e.g
            'sqlite:///path/to/file.db::table'
        imports : list
            A list of string names for any modules that must be imported on
            the Blaze server before the resource can be added. This is identical
            to the `imports` field in a Blaze server YAML file.
        args : any, optional
            Any additional positional arguments that can be passed to the
            ``blaze.resource`` constructor for this resource type
        kwargs : any, optional
            Any additional keyword arguments that can be passed to the
            ``blaze.resource`` constructor for this resource type
        """
        payload = {name: {'source': resource_uri}}
        imports = kwargs.pop('imports', None)
        if imports is not None:
            payload[name]['imports'] = imports
        if args:
            payload[name]['args'] = args
        if kwargs:
            payload[name]['kwargs'] = kwargs

        response = post(self, '/add', auth=self.auth,
                        data=self.serial.dumps(payload),
                        headers={'Content-Type': 'application/vnd.blaze+' + self.serial.name})
        # A special case for the "Not Found" error, since that means that this
        # server doesn't support adding datasets, and the user should see a more
        # helpful response
        if response.status_code == 404:
            raise ValueError("Server does not support dynamically adding datasets")

        if not ok(response):
            raise ValueError("Bad Response: %s" % reason(response))


@dispatch(Client)
def discover(c):
    return c.dshape


def mimetype(serial):
    """Function to generate a blaze serialization format mimetype put into a
    dictionary of headers for consumption by requests.

    Examples
    --------
    >>> from blaze.server.serialization import msgpack
    >>> mimetype(msgpack)
    {'Content-Type': 'application/vnd.blaze+msgpack'}
    """
    return {'Content-Type': 'application/vnd.blaze+%s' % serial.name}


@dispatch(Expr, Client)
def compute_down(expr,
                 ec,
                 profiler_output=None,
                 compute_kwargs=None,
                 odo_kwargs=None,
                 **kwargs):
    """Compute down for blaze clients.

    Parameters
    ----------
    expr : Expr
        The expression to send to the server.
    ec : Client
        The blaze client to compute against.
    namespace : dict[Symbol -> any], optional
        The namespace to compute the expression in. This will be amended to
        include that data for the server. By default this will just be the
        client mapping to the server's data.
    compute_kwargs : dict, optional
        Extra kwargs to pass to compute on the server.
    odo_kwargs : dict, optional
        Extra kwargs to pass to odo on the server.
    profile : bool, optional
        Should blaze server run cProfile over the computation of the expression
        and the serialization of the response.
    profiler_output : file-like object, optional
        A file like object to hold the profiling output from the server.
        If this is not passed then the server will write the data to the
        server's filesystem
    """
    from .server import to_tree

    kwargs = keymap(u8, kwargs)

    tree = to_tree(expr)
    serial = ec.serial
    if profiler_output is not None:
        kwargs[u'profile'] = True
        kwargs[u'profiler_output'] = ':response'

    kwargs[u'compute_kwargs'] = keymap(u8, compute_kwargs or {})
    kwargs[u'odo_kwargs'] = keymap(u8, odo_kwargs or {})

    r = post(
        ec,
        '/compute',
        data=serial.dumps(assoc(kwargs, u'expr', tree)),
        auth=ec.auth,
        headers=mimetype(serial),
    )

    if not ok(r):
        raise ValueError("Bad response: %s" % reason(r))
    response = serial.loads(content(r))
    if profiler_output is not None:
        profiler_output.write(response[u'profiler_output'])
    return serial.data_loads(response[u'data'])


@resource.register('blaze://.+')
def resource_blaze(uri, leaf=None, **kwargs):
    if leaf is not None:
        raise ValueError('The syntax blaze://...::{leaf} is no longer '
                         'supported as of version 0.8.1.\n'
                         'You can access {leaf!r} using this syntax:\n'
                         'data({uri})[{leaf!r}]'
                         .format(leaf=leaf, uri=uri))
    uri = uri[len('blaze://'):]
    sp = uri.split('/')
    tld, rest = sp[0], sp[1:]
    if ':' not in tld:
        tld += ':%d' % DEFAULT_PORT
    uri = '/'.join([tld] + list(rest))
    return Client(uri)



from __future__ import absolute_import, division, print_function

from datetime import datetime
from decimal import Decimal

import sqlalchemy as sa
import sqlalchemy.orm
from toolz import curry
from datashape.predicates import isrecord
from ..expr import Field
from odo.backends.sql import dshape_to_alchemy

# This was taken from the following StackOverflow post
# http://stackoverflow.com/questions/5631078/sqlalchemy-print-the-actual-query
# answer by bukzor http://stackoverflow.com/users/146821/bukzor


def literalquery(statement, dialect=None):
    """Generate an SQL expression string with bound parameters rendered inline
    for the given SQLAlchemy statement.

    WARNING: This method of escaping is insecure, incomplete, and for debugging
    purposes only. Executing SQL statements with inline-rendered user values is
    extremely insecure.
    """
    if isinstance(statement, sqlalchemy.orm.Query):
        if dialect is None:
            dialect = statement.session.get_bind(
                statement._mapper_zero_or_none()
            ).dialect
        statement = statement.statement
    if dialect is None:
        dialect = getattr(statement.bind, 'dialect', None)
    if dialect is None:
        from sqlalchemy.dialects import mysql
        dialect = mysql.dialect()

    Compiler = type(statement._compiler(dialect))

    class LiteralCompiler(Compiler):
        visit_bindparam = Compiler.render_literal_bindparam

        def render_literal_value(self, value, type_):
            if isinstance(value, (Decimal, long)):
                return str(value)
            elif isinstance(value, datetime):
                return repr(str(value))
            else:  # fallback
                value = super(LiteralCompiler, self).render_literal_value(
                    value, type_,
                )
                if isinstance(value, unicode):
                    return value.encode('UTF-8')
                else:
                    return value

    return LiteralCompiler(dialect, statement)


def make_sqlalchemy_table(expr):
    return sa.Table(expr._name, sa.MetaData(), *dshape_to_alchemy(expr.dshape))


@curry
def istable(db, t):
    return (isinstance(t, Field) and
            isrecord(t.dshape.measure) and
            t._child.isidentical(db))



"""SparkSQL backend for blaze.

Notes
-----
Translation happens via the Hive sqlalchemy dialect, which is then sent to
SparkSQL.
"""

from __future__ import absolute_import, division, print_function

from functools import reduce
from operator import and_
from distutils.version import LooseVersion

from toolz import pipe
from toolz.curried import filter, map

from sqlalchemy.ext.compiler import compiles
import sqlalchemy as sa

from ..dispatch import dispatch
from ..expr import Expr, symbol, Join
from .core import compute
from .utils import literalquery, istable, make_sqlalchemy_table
from ..utils import listpack
from .spark import jgetattr

from pyhive.sqlalchemy_hive import HiveDialect

from pyspark import SQLContext
from pyspark.sql import DataFrame as SparkDataFrame

__all__ = []

join_types = {
    'left': 'left_outer',
    'right': 'right_outer'
}


@dispatch(Join, SparkDataFrame, SparkDataFrame)
def compute_up(t, lhs, rhs, **kwargs):
    ands = [getattr(lhs, left) == getattr(rhs, right)
            for left, right in zip(*map(listpack, (t.on_left, t.on_right)))]

    joined = lhs.join(rhs, reduce(and_, ands), join_types.get(t.how, t.how))

    prec, sec = (rhs, lhs) if t.how == 'right' else (lhs, rhs)
    cols = [jgetattr(prec, f, jgetattr(sec, f, None)) for f in t.fields]
    assert all(c is not None for c in cols)
    return joined.select(*cols)


if LooseVersion(sa.__version__) >= '1.0.0':
    # a bug in spark sql prevents labels from being referenced properly
    @compiles(sa.sql.elements._label_reference, 'hive')
    def compile_label_reference(element, compiler, **kwargs):
        return compiler.process(element.element, **kwargs)


@dispatch(Expr, SQLContext)
def compute_down(expr, data, **kwargs):
    """ Compile a blaze expression to a sparksql expression"""
    leaves = expr._leaves()

    # make sure we only have a single leaf node
    if len(leaves) != 1:
        raise ValueError('Must compile from exactly one root database')

    leaf, = leaves

    # field expressions on the database are Field instances with a record
    # measure whose immediate child is the database leaf
    tables = pipe(expr._subterms(), filter(istable(leaf)), list)

    # raise if we don't have tables in our database
    if not tables:
        raise ValueError('Expressions not referencing a table cannot be '
                         'compiled')

    # make new symbols for each table
    new_leaves = [symbol(t._name, t.dshape) for t in tables]

    # sub them in the expression
    expr = expr._subs(dict(zip(tables, new_leaves)))

    # compute using sqlalchemy
    scope = dict(zip(new_leaves, map(make_sqlalchemy_table, tables)))
    query = compute(expr, scope, return_type='native')

    # interpolate params
    compiled = literalquery(query, dialect=HiveDialect())
    return data.sql(str(compiled))



from __future__ import absolute_import, division, print_function

from numbers import Number
from toolz import concat, first, curry, compose
from datashape import DataShape

from blaze import compute, ndim
from blaze.dispatch import dispatch
from blaze.compute.core import compute_up, optimize
from blaze.expr import (ElemWise, symbol, Reduction, Transpose, TensorDot,
                        Expr, Slice, Broadcast)
from blaze.expr.split import split

from dask.array.core import _concatenate2, Array, atop, transpose, tensordot


def compute_it(expr, leaves, *data, **kwargs):
    kwargs.pop('scope')
    return compute(expr, dict(zip(leaves, data)), **kwargs)


def elemwise_array(expr, *data, **kwargs):
    leaves = expr._inputs
    expr_inds = tuple(range(ndim(expr)))[::-1]
    return atop(curry(compute_it, expr, leaves, **kwargs),
                expr_inds,
                *concat((dat, tuple(range(ndim(dat))[::-1])) for dat in data))


try:
    from blaze.compute.numba import (get_numba_ufunc, broadcast_collect,
            Broadcastable)

    def compute_broadcast(expr, *data, **kwargs):
        expr_inds = tuple(range(ndim(expr)))[::-1]
        func = get_numba_ufunc(expr)
        return atop(func,
                    expr_inds,
                    *concat((dat, tuple(range(ndim(dat))[::-1])) for dat in data))

    def optimize_array(expr, *data):
        return broadcast_collect(
            expr,
            broadcastable=Broadcastable,
            want_to_broadcast=Broadcastable,
        )

    for i in range(5):
        compute_up.register(Broadcast, *([(Array, Number)] * i))(compute_broadcast)
        optimize.register(Expr, *([(Array, Number)] * i))(optimize_array)

except ImportError:
    pass


for i in range(5):
    compute_up.register(ElemWise, *([Array] * i))(elemwise_array)


@dispatch(Reduction, Array)
def compute_up(expr, data, **kwargs):
    leaf = expr._leaves()[0]
    chunk = symbol('chunk', DataShape(*(tuple(map(first, data.chunks)) +
                                        (leaf.dshape.measure,))))
    (chunk, chunk_expr), (agg, agg_expr) = split(expr._child, expr,
                                                 chunk=chunk)

    inds = tuple(range(ndim(leaf)))
    tmp = atop(curry(compute_it, chunk_expr, [chunk], **kwargs), inds, data,
               inds)

    return atop(compose(curry(compute_it, agg_expr, [agg], **kwargs),
                        curry(_concatenate2, axes=expr.axis)),
                tuple(i for i in inds if i not in expr.axis),
                tmp, inds)


@dispatch(Transpose, Array)
def compute_up(expr, data, **kwargs):
    return transpose(data, expr.axes)


@dispatch(TensorDot, Array, Array)
def compute_up(expr, lhs, rhs, **kwargs):
    return tensordot(lhs, rhs, (expr._left_axes, expr._right_axes))


@dispatch(Slice, Array)
def compute_up(expr, data, **kwargs):
    return data[expr.index]



from __future__ import absolute_import, division, print_function

import pytest
pytest.importorskip('flask')

from pandas import DataFrame
from blaze import compute, by, into, discover
from blaze import data as bz_data
from blaze.expr import Expr, symbol, Field
from blaze.dispatch import dispatch
from blaze.server import Server
from blaze.server.client import Client
from blaze.utils import example


df = DataFrame([['Alice', 100], ['Bob', 200]],
               columns=['name', 'amount'])

df2 = DataFrame([['Charlie', 100], ['Dan', 200]],
                columns=['name', 'amount'])

tdata = {'accounts': df, 'accounts2': df}

server = Server(tdata)
add_server = Server(tdata, allow_add=True)

test = server.app.test_client()
test_add = add_server.app.test_client()

from blaze.server import client
client.requests = test # OMG monkey patching


def test_client():
    c = Client('localhost:6363')
    assert str(discover(c)) == str(discover(tdata))

    t = symbol('t', discover(c))
    expr = t.accounts.amount.sum()

    assert compute(expr, c) == 300
    assert 'name' in t.accounts.fields
    assert isinstance(t.accounts.name, Field)
    assert compute(t.accounts.name, c) == ['Alice', 'Bob']


def test_expr_client_interactive():
    c = Client('localhost:6363')
    t = bz_data(c)

    assert compute(t.accounts.name) == ['Alice', 'Bob']
    assert (into(set, compute(by(t.accounts.name, min=t.accounts.amount.min(),
                                                  max=t.accounts.amount.max())))
            == set([('Alice', 100, 100), ('Bob', 200, 200)]))


def test_compute_client_with_multiple_datasets():
    c = bz_data('blaze://localhost:6363')
    s = symbol('s', discover(c))

    assert compute(s.accounts.amount.sum() + s.accounts2.amount.sum(),
                    {s: c}) == 600


def test_bz_data():
    c = bz_data('blaze://localhost:6363')
    assert isinstance(c.data, Client)
    assert str(discover(c)) == str(discover(tdata))


def test_bz_data_default_port():
    ec = bz_data('blaze://localhost')
    assert str(discover(ec)) == str(discover(tdata))


def test_bz_data_non_default_port():
    ec = bz_data('blaze://localhost:6364')
    assert ec.data.url == 'http://localhost:6364'


def test_bz_data_all_in_one():
    ec = bz_data('blaze://localhost:6363')
    assert str(discover(ec)) == str(discover(tdata))


class CustomExpr(Expr):
    __slots__ = '_hash', '_child'

    @property
    def dshape(self):
        return self._child.dshape


@dispatch(CustomExpr, DataFrame)
def compute_up(expr, tdata, **kwargs):
    return tdata


def test_custom_expressions():
    ec = Client('localhost:6363')
    t = symbol('t', discover(ec))

    assert list(map(tuple, compute(CustomExpr(t.accounts), ec))) == into(list, df)


def test_client_dataset_fails():
    with pytest.raises(ValueError):
        bz_data('blaze://localhost::accounts')
    with pytest.raises(ValueError):
        bz_data('blaze://localhost::accounts')


def test_client_dataset():
    d = bz_data('blaze://localhost')
    assert list(map(tuple, into(list, d.accounts))) == into(list, df)


def test_client_cant_add_dataset():
    ec = Client('localhost:6363')
    with pytest.raises(ValueError) as excinfo:
        ec.add('iris', example('iris.csv'))
    assert "Server does not support" in str(excinfo.value)


def test_client_add_dataset():
    client.requests = test_add  # OMG more monkey patching
    ec = Client('localhost:6363')
    ec.add('iris', example('iris.csv'))
    assert 'iris' in ec.dshape.measure.dict
    iris_data = bz_data(example('iris.csv'))
    assert ec.dshape.measure.dict['iris'] == iris_data.dshape


def test_client_add_dataset_failure():
    client.requests = test_add  # OMG more monkey patching
    ec = Client('localhost:6363')
    with pytest.raises(ValueError) as exc:
        ec.add('iris2', example('iris.csv'), -1, bad_arg='value')
    assert '422 UNPROCESSABLE ENTITY' in str(exc.value)


def test_client_add_dataset_with_args():
    client.requests = test_add  # OMG more monkey patching
    ec = Client('localhost:6363')
    ec.add('teams', 'sqlite:///' + example('teams.db'), 'teams',
           primary_key='teamID')
    assert 'teams' in ec.dshape.measure.dict
    teams_data = bz_data('sqlite:///' + example('teams.db') + '::teams')
    assert ec.dshape.measure.dict['teams'] == teams_data.dshape



from itertools import product
import pytest

from blaze.expr import Add, USub, Not, Gt, Mult, Concat, Interp, Repeat, \
    Relational
from blaze import symbol
from datashape import dshape, Option


x = symbol('x', '5 * 3 * int32')
y = symbol('y', '5 * 3 * int32')
w = symbol('w', '5 * 3 * float32')
z = symbol('z', '5 * 3 * int64')
a = symbol('a', 'int32')
b = symbol('b', '5 * 3 * bool')
cs = symbol('cs', 'string')
d = symbol('d', '5*3*?int32')
e = symbol('e', '?int32')
f = symbol('f', '?bool')
g = symbol('g', 'datetime')
h = symbol('h', 'datetime')
i = symbol('i', '?datetime')
j = symbol('j', '?datetime')
k = symbol('k', 'timedelta')
l = symbol('l', '?timedelta')
optionals = {d, e, f}


def test_arithmetic_dshape_on_collections():
    assert Add(x, y).shape == x.shape == y.shape


def test_arithmetic_broadcasts_to_scalars():
    assert Add(x, a).shape == x.shape
    assert Add(x, 1).shape == x.shape


def test_unary_ops_are_elemwise():
    assert USub(x).shape == x.shape
    assert Not(b).shape == b.shape


@pytest.mark.parametrize('sym', (b, f))
def test_not_dshape(sym):
    if sym in optionals:
        assert Not(sym).schema == dshape(Option('bool'))
    else:
        assert Not(sym).schema == dshape('bool')


@pytest.mark.parametrize('relation_type', Relational.__subclasses__())
def test_relations_maintain_shape(relation_type):
    assert relation_type(x, y).shape == x.shape


@pytest.mark.parametrize('relation_type', Relational.__subclasses__())
@pytest.mark.parametrize('lhs,rhs', product((x, y, d, e, f), repeat=2))
def test_relations_are_boolean(lhs, rhs, relation_type):
    if lhs in optionals or rhs in optionals:
        assert relation_type(lhs, rhs).schema == dshape(Option('bool'))
    else:
        assert relation_type(lhs, rhs).schema == dshape('bool')


def test_names():
    assert Add(x, 1)._name == x._name
    assert Add(1, x)._name == x._name
    assert Mult(Add(1, x), 2)._name == x._name

    assert Add(y, x)._name != x._name
    assert Add(y, x)._name != y._name

    assert Add(x, x)._name == x._name


def test_inputs():
    assert (x + y)._inputs == (x, y)
    assert (x + 1)._inputs == (x,)
    assert (1 + y)._inputs == (y,)


def test_printing():
    assert str(-x) == '-x'
    assert str(-(x + y)) == '-(x + y)'

    assert str(~b) == '~b'
    assert str(~(b | (x > y))) == '~(b | (x > y))'


def test_dir():
    i = symbol('i', '10 * int')
    d = symbol('d', '10 * datetime')

    assert isinstance(i + 1, Add)  # this works
    with pytest.raises(Exception):  # this doesn't
        d + 1


def test_arith_ops_promote_dtype():
    r = w + z
    assert r.dshape == dshape('5 * 3 * float64')


def test_str_arith():
    rep = cs * 1
    assert isinstance(rep, Repeat)
    assert rep.lhs.isidentical(cs)
    assert rep.rhs == 1

    with pytest.raises(TypeError):
        cs * 1.5

    with pytest.raises(TypeError):
        cs * 'a'

    interp = cs % cs
    assert isinstance(interp, Interp)
    assert interp.lhs.isidentical(cs)
    assert interp.rhs.isidentical(cs)

    interp_lit = cs % 'a'
    assert isinstance(interp_lit, Interp)
    assert interp_lit.lhs.isidentical(cs)
    assert interp_lit.rhs == 'a'

    with pytest.raises(Exception):
        cs / 1

    with pytest.raises(Exception):
        cs // 1


def test_datetime_sub():
    assert (g - h).dshape == dshape('timedelta')
    assert (g - i).dshape == dshape('?timedelta')
    assert (i - j).dshape == dshape('?timedelta')
    assert (g - k).dshape == dshape('datetime')
    assert (g - l).dshape == dshape('?datetime')
    assert (i - k).dshape == dshape('?datetime')
    assert (i - l).dshape == dshape('?datetime')


def test_datetime_add():
    assert (g + k).dshape == dshape('datetime')
    assert (g + l).dshape == dshape('?datetime')
    assert (i + k).dshape == dshape('?datetime')
    assert (i + l).dshape == dshape('?datetime')



from __future__ import absolute_import, division, print_function

import pytest
pymongo = pytest.importorskip('pymongo')

from datetime import datetime
from toolz import pluck, reduceby, groupby

from datashape import Record
import blaze
from blaze import compute, compute_up, discover, dshape, data

from blaze.compute.mongo import MongoQuery
from blaze.expr import symbol, by, floor, ceil
from blaze.compatibility import xfail


def into(*args, **kwargs):
    try:
        return blaze.into(*args, **kwargs)
    except pymongo.errors.ConnectionFailure:
        pytest.skip('No mongo server running')


@pytest.fixture(scope='module')
def mongo_host_port():
    import os
    return (os.environ.get('MONGO_IP', 'localhost'),
            os.environ.get('MONGO_PORT', 27017))


@pytest.fixture(scope='module')
def conn(mongo_host_port):
    host, port = mongo_host_port
    try:
        return pymongo.MongoClient(host=host, port=port)
    except pymongo.errors.ConnectionFailure:
        pytest.skip('No mongo server running')


@pytest.fixture(scope='module')
def db(conn):
    return conn.test_db


bank_raw = [{'name': 'Alice', 'amount': 100},
            {'name': 'Alice', 'amount': 200},
            {'name': 'Bob', 'amount': 100},
            {'name': 'Bob', 'amount': 200},
            {'name': 'Bob', 'amount': 300}]


@pytest.yield_fixture
def big_bank(db):
    data = [{'name': 'Alice', 'amount': 100, 'city': 'New York City'},
            {'name': 'Alice', 'amount': 200, 'city': 'Austin'},
            {'name': 'Bob', 'amount': 100, 'city': 'New York City'},
            {'name': 'Bob', 'amount': 200, 'city': 'New York City'},
            {'name': 'Bob', 'amount': 300, 'city': 'San Francisco'}]
    coll = db.bigbank
    coll = into(coll, data)
    try:
        yield coll
    finally:
        coll.drop()


@pytest.yield_fixture
def date_data(db):
    n = 3
    d = {'name': ['Alice', 'Bob', 'Joe'],
         'when': [datetime(2010, 1, 1, i) for i in [1, 2, 3]],
         'amount': [100, 200, 300],
         'id': [1, 2, 3]}
    data = [dict(zip(d.keys(), [d[k][i] for k in d.keys()]))
            for i in range(n)]
    coll = into(db.date_data, data)
    try:
        yield coll
    finally:
        coll.drop()


@pytest.yield_fixture
def bank(db):
    coll = db.bank
    coll = into(coll, bank_raw)
    try:
        yield coll
    finally:
        coll.drop()


@pytest.yield_fixture
def missing_vals(db):
    data = [{'x': 1, 'z': 100},
            {'x': 2, 'y': 20, 'z': 200},
            {'x': 3, 'z': 300},
            {'x': 4, 'y': 40}]
    coll = db.missing_vals
    coll = into(coll, data)
    try:
        yield coll
    finally:
        coll.drop()


@pytest.yield_fixture
def points(db):
    data = [{'x': 1, 'y': 10, 'z': 100},
            {'x': 2, 'y': 20, 'z': 200},
            {'x': 3, 'y': 30, 'z': 300},
            {'x': 4, 'y': 40, 'z': 400}]
    coll = db.points
    coll = into(coll, data)
    try:
        yield coll
    finally:
        coll.drop()


@pytest.yield_fixture
def events(db):
    data = [{'time': datetime(2012, 1, 1, 12, 00, 00), 'x': 1},
            {'time': datetime(2012, 1, 2, 12, 00, 00), 'x': 2},
            {'time': datetime(2012, 1, 3, 12, 00, 00), 'x': 3}]
    coll = db.events
    coll = into(coll, data)
    try:
        yield coll
    finally:
        coll.drop()


t = symbol('t', 'var * {name: string, amount: int}')
bigt = symbol('bigt', 'var * {name: string, amount: int, city: string}')
p = symbol('p', 'var * {x: int, y: int, z: int}')
e = symbol('e', 'var * {time: datetime, x: int}')


q = MongoQuery('fake', [])


def test_compute_on_db(bank, points):
    assert bank.database == points.database
    db = bank.database

    d = symbol(db.name, discover(db))
    assert (compute(d.points.x.sum(), db) ==
            sum(x['x'] for x in db.points.find()))


def test_symbol(bank):
    assert compute(t, bank) == list(pluck(['name', 'amount'], bank_raw))


def test_projection_one():
    assert compute_up(t[['name']], q).query == ({'$project': {'name': 1}},)


def test_head_one():
    assert compute_up(t.head(5), q).query == ({'$limit': 5},)


def test_head(bank):
    assert len(compute(t.head(2), bank)) == 2


def test_projection(bank):
    assert set(compute(t.name, bank)) == set(['Alice', 'Bob'])
    assert set(compute(t[['name']], bank)) == set([('Alice',), ('Bob',)])


def test_selection(bank):
    assert set(compute(t[t.name == 'Alice'], bank)) == set([('Alice', 100),
                                                            ('Alice', 200)])
    assert set(compute(t['Alice' == t.name], bank)) == set([('Alice', 100),
                                                            ('Alice', 200)])
    assert set(compute(t[t.amount > 200], bank)) == set([('Bob', 300)])
    assert set(compute(t[t.amount >= 200], bank)) == set([('Bob', 300),
                                                          ('Bob', 200),
                                                          ('Alice', 200)])
    assert set(compute(t[t.name != 'Alice'].name, bank)) == set(['Bob'])
    assert set(compute(t[(t.name == 'Alice') & (t.amount > 150)], bank)) == \
        set([('Alice', 200)])
    assert set(compute(t[(t.name == 'Alice') | (t.amount > 250)], bank)) == \
        set([('Alice', 200),
             ('Alice', 100),
             ('Bob', 300)])


def test_columnwise(points):
    assert set(compute(p.x + p.y, points)) == set([11, 22, 33, 44])


def test_columnwise_multiple_operands(points):
    expected = [x['x'] + x['y'] - x['z'] * x['x'] / 2 for x in points.find()]
    assert set(compute(p.x + p.y - p.z * p.x / 2, points)) == set(expected)


def test_arithmetic(points):
    expr = p.y // p.x
    assert set(compute(expr, points)) == set(compute(expr, points.find()))


def test_columnwise_mod(points):
    expected = [x['x'] % x['y'] - x['z'] * x['x'] / 2 + 1
                for x in points.find()]
    expr = p.x % p.y - p.z * p.x / 2 + 1
    assert set(compute(expr, points)) == set(expected)


@xfail(raises=NotImplementedError,
       reason='MongoDB does not implement certain arith ops')
def test_columnwise_pow(points):
    expected = [x['x'] ** x['y'] for x in points.find()]
    assert set(compute(p.x ** p.y, points)) == set(expected)


def test_by_one():
    assert compute_up(by(t.name, total=t.amount.sum()), q).query == \
        ({'$group': {'_id': {'name': '$name'},
                     'total': {'$sum': '$amount'}}},
         {'$project': {'total': '$total', 'name': '$_id.name'}})


def test_by(bank):
    assert set(compute(by(t.name, total=t.amount.sum()), bank)) == \
        set([('Alice', 300), ('Bob', 600)])
    assert set(compute(by(t.name, min=t.amount.min()), bank)) == \
        set([('Alice', 100), ('Bob', 100)])
    assert set(compute(by(t.name, max=t.amount.max()), bank)) == \
        set([('Alice', 200), ('Bob', 300)])
    assert set(compute(by(t.name, count=t.name.count()), bank)) == \
        set([('Alice', 2), ('Bob', 3)])


def test_reductions(bank):
    assert compute(t.amount.min(), bank) == 100
    assert compute(t.amount.max(), bank) == 300
    assert compute(t.amount.sum(), bank) == 900


def test_distinct(bank):
    assert set(compute(t.name.distinct(), bank)) == set(['Alice', 'Bob'])


def test_nunique_collection(bank):
    assert compute(t.nunique(), bank) == len(bank_raw)


def test_sort(bank):
    assert compute(t.amount.sort('amount'), bank) == \
        [100, 100, 200, 200, 300]
    assert compute(t.amount.sort('amount', ascending=False), bank) == \
        [300, 200, 200, 100, 100]


def test_by_multi_column(bank):
    assert set(compute(by(t[['name', 'amount']], count=t.count()), bank)) == \
        set([(d['name'], d['amount'], 1) for d in bank_raw])


def test_datetime_handling(events):
    assert set(compute(e[e.time >= datetime(2012, 1, 2, 12, 0, 0)].x,
                       events)) == set([2, 3])
    assert set(compute(e[e.time >= "2012-01-02"].x,
                       events)) == set([2, 3])


def test_summary_kwargs(bank):
    expr = by(t.name, total=t.amount.sum(), avg=t.amount.mean())
    result = compute(expr, bank)
    assert result == [('Bob', 200.0, 600), ('Alice', 150.0, 300)]


def test_summary_count(bank):
    expr = by(t.name, how_many=t.amount.count())
    result = compute(expr, bank)
    assert result == [('Bob', 3), ('Alice', 2)]


def test_summary_arith(bank):
    expr = by(t.name, add_one_and_sum=(t.amount + 1).sum())
    result = compute(expr, bank)
    assert result == [('Bob', 603), ('Alice', 302)]


def test_summary_arith_min(bank):
    expr = by(t.name, add_one_and_sum=(t.amount + 1).min())
    result = compute(expr, bank)
    assert result == [('Bob', 101), ('Alice', 101)]


def test_summary_arith_max(bank):
    expr = by(t.name, add_one_and_sum=(t.amount + 1).max())
    result = compute(expr, bank)
    assert result == [('Bob', 301), ('Alice', 201)]


def test_summary_complex_arith(bank):
    expr = by(t.name, arith=(100 - t.amount * 2 / 30.0).sum())
    result = compute(expr, bank)
    reducer = lambda acc, x: (100 - x['amount'] * 2 / 30.0) + acc
    expected = reduceby('name', reducer, bank.find(), 0)
    assert set(result) == set(expected.items())


def test_summary_complex_arith_multiple(bank):
    expr = by(t.name, arith=(100 - t.amount * 2 / 30.0).sum(),
              other=t.amount.mean())
    result = compute(expr, bank)
    reducer = lambda acc, x: (100 - x['amount'] * 2 / 30.0) + acc
    expected = reduceby('name', reducer, bank.find(), 0)

    mu = reduceby('name', lambda acc, x: acc + x['amount'], bank.find(), 0.0)
    values = list(mu.values())
    items = expected.items()
    counts = groupby('name', bank.find())
    items = [x + (float(v) / len(counts[x[0]]),)
             for x, v in zip(items, values)]
    assert set(result) == set(items)


def test_like(bank):
    bank.create_index([('name', pymongo.TEXT)])
    expr = t[t.name.like('*Alice*')]
    result = compute(expr, bank)
    assert set(result) == set((('Alice', 100), ('Alice', 200)))


def test_like_multiple(big_bank):
    expr = bigt[bigt.name.like('*Bob*') & bigt.city.like('*York*')]
    result = compute(expr, big_bank)
    assert set(result) == set(
        (('Bob', 100, 'New York City'), ('Bob', 200, 'New York City'))
    )


def test_like_mulitple_no_match(big_bank):
    # make sure we aren't OR-ing the matches
    expr = bigt[bigt.name.like('*York*') & bigt.city.like('*Bob*')]
    assert not set(compute(expr, big_bank))


def test_missing_values(missing_vals):
    assert discover(missing_vals).subshape[0] == \
        dshape('{x: int64, y: ?int64, z: ?int64}')

    assert set(compute(p.y, missing_vals)) == set([None, 20, None, 40])


def test_datetime_access(date_data):
    t = symbol('t',
               ('var * {amount: float64, id: int64,'
                      ' name: string, when: datetime}'))

    py_data = into(list, date_data)  # a python version of the collection

    for attr in ['day', 'minute', 'second', 'year', 'month']:
        assert list(compute(getattr(t.when, attr), date_data)) == \
            list(compute(getattr(t.when, attr), py_data))


def test_datetime_access_and_arithmetic(date_data):
    t = symbol('t',
               ('var * {amount: float64, id: int64,'
                       ' name: string, when: datetime}'))

    py_data = into(list, date_data)  # a python version of the collection

    expr = t.when.day + t.id

    assert list(compute(expr, date_data)) == list(compute(expr, py_data))


def test_floor_ceil(bank):
    t = symbol('t', discover(bank))
    assert set(compute(200 * floor(t.amount / 200), bank)) == set([0, 200])
    assert set(compute(200 * ceil(t.amount / 200), bank)) == set([200, 400])


def test_Data_construct(bank, points, mongo_host_port):
    d = data('mongodb://{}:{}/test_db'.format(*mongo_host_port))
    assert 'bank' in d.fields
    assert 'points' in d.fields
    assert isinstance(d.dshape.measure, Record)


def test_Data_construct_with_table(bank, mongo_host_port):
    d = data('mongodb://{}:{}/test_db::bank'.format(*mongo_host_port))
    assert set(d.fields) == set(('name', 'amount'))
    assert int(d.count()) == 5


def test_and_same_key(bank):
    expr = t[(t.amount > 100) & (t.amount < 300)]
    result = compute(expr, bank)
    expected = [('Alice', 200), ('Bob', 200)]
    assert result == expected


def test_interactive_dshape_works(bank, mongo_host_port):
    try:
        d = data('mongodb://{}:{}/test_db::bank'.format(*mongo_host_port))
    except pymongo.errors.ConnectionFailure:
        pytest.skip('No mongo server running')
    assert dshape(d.dshape.measure) == dshape('{amount: int64, name: string}')


@pytest.mark.xfail(raises=TypeError, reason="IsIn not yet implemented")
def test_isin_fails(bank):
    expr = t[t.amount.isin([100])]
    result = compute(expr, bank)
    assert result == compute(t[t.amount == 100], bank)



from __future__ import absolute_import, division, print_function

import warnings

import datashape
from datashape import String, DataShape, Option, bool_
from odo.utils import copydoc

from .expressions import schema_method_list, ElemWise
from .arithmetic import Interp, Repeat, _mkbin, repeat, interp, _add, _radd
from ..compatibility import basestring, _inttypes, builtins
from ..deprecation import deprecated

__all__ = ['Like',
           'like',
           'Pad',
           'Replace',
           'SliceReplace',
           # prevent 'len' to end up in global namespace
           #'len',
           'upper',
           'lower',
           'cat',
           'isalnum',
           'isalpha',
           'isdecimal',
           'isdigit',
           'islower',
           'isnumeric',
           'isspace',
           'istitle',
           'isupper',
           'StrCat',
           'find',
           'StrFind',
           'StrSlice',
           'slice',
           'slice_replace',
           'replace',
           'capitalize',
           'strip',
           'lstrip',
           'rstrip',
           'pad',
           'UnaryStringFunction']

def _validate(var, name, type, typename):
    if not isinstance(var, type):
        raise TypeError('"%s" argument must be a %s'%(name, typename))

def _validate_optional(var, name, type, typename):
    if var is not None and not isinstance(var, type):
        raise TypeError('"%s" argument must be a %s'%(name, typename))



class Like(ElemWise):

    """ Filter expression by string comparison

    >>> from blaze import symbol, like, compute
    >>> t = symbol('t', 'var * {name: string, city: string}')
    >>> expr = t[t.name.like('Alice*')]

    >>> data = [('Alice Smith', 'New York'),
    ...         ('Bob Jones', 'Chicago'),
    ...         ('Alice Walker', 'LA')]
    >>> list(compute(expr, data))
    [('Alice Smith', 'New York'), ('Alice Walker', 'LA')]
    """
    _arguments = '_child', 'pattern'

    def _dshape(self):
        shape, schema = self._child.dshape.shape, self._child.schema
        schema = Option(bool_) if isinstance(schema.measure, Option) else bool_
        return DataShape(*(shape + (schema,)))


@copydoc(Like)
def like(child, pattern):
    if not isinstance(pattern, basestring):
        raise TypeError('pattern argument must be a string')
    return Like(child, pattern)


class UnaryStringFunction(ElemWise):

    """String function that only takes a single argument.
    """
    _arguments = '_child',


class len(UnaryStringFunction):
    schema = datashape.int64


class upper(UnaryStringFunction):

    @property
    def schema(self):
        return self._child.schema


class lower(UnaryStringFunction):

    @property
    def schema(self):
        return self._child.schema


class PredicateFunction(UnaryStringFunction):

    @property
    def schema(self):
        return bool_ if self._child.schema == datashape.string else Option(bool_)


class isalnum(PredicateFunction): pass
class isalpha(PredicateFunction): pass
class isdecimal(PredicateFunction): pass
class isdigit(PredicateFunction): pass
class islower(PredicateFunction): pass
class isnumeric(PredicateFunction): pass
class isspace(PredicateFunction): pass
class istitle(PredicateFunction): pass
class isupper(PredicateFunction): pass

class StrFind(ElemWise):
    """
    Find literal substring in string column.

    """

    _arguments = '_child', 'sub'
    schema = Option(datashape.int64)


@copydoc(StrFind)
def find(col, sub):
    if not isinstance(sub, basestring):
        raise TypeError("'sub' argument must be a String")
    return StrFind(col, sub)

class Replace(ElemWise):
    _arguments = '_child', 'old', 'new', 'max'

    @property
    def schema(self):
        return self._child.schema

def replace(col, old, new, max=None):
    _validate(old, 'old', basestring, 'string')
    _validate(new, 'new', basestring, 'string')
    _validate_optional(max, 'max', int, 'integer')
    return Replace(col, old, new, max)

class Pad(ElemWise):
    _arguments = '_child', 'width', 'side', 'fillchar'

    @property
    def schema(self):
        return self._child.schema

def pad(col, width, side=None, fillchar=None):
    _validate(width, 'width', int, 'integer')
    if side not in (None, 'left', 'right'):
        raise TypeError('"side" argument must be either "left" or "right"')
    _validate_optional(fillchar, 'fillchar', basestring, 'string')
    return Pad(col, width, side, fillchar)

class capitalize(UnaryStringFunction):

    @property
    def schema(self):
        return self._child.schema

class strip(UnaryStringFunction):

    @property
    def schema(self):
        return self._child.schema

class lstrip(UnaryStringFunction):

    @property
    def schema(self):
        return self._child.schema

class rstrip(UnaryStringFunction):

    @property
    def schema(self):
        return self._child.schema

class StrSlice(ElemWise):
    _arguments = '_child', 'slice'

    @property
    def schema(self):
        return self._child.schema

class SliceReplace(ElemWise):
    _arguments = '_child', 'start', 'stop', 'repl'

    @property
    def schema(self):
        return self._child.schema

def slice_replace(col, start=None, stop=None, repl=None):
    _validate_optional(start, 'start', int, 'integer')
    _validate_optional(stop, 'stop', int, 'integer')
    _validate_optional(repl, 'repl', basestring, 'string')
    return SliceReplace(col, start, stop, repl)


@copydoc(StrSlice)
def slice(col, idx):
    if not isinstance(idx, (builtins.slice, _inttypes)):
        raise TypeError("idx argument must be a slice or integer, given {}".format(slc))
    return StrSlice(col, (idx.start, idx.stop, idx.step) if isinstance(idx, builtins.slice) else idx)

class StrCat(ElemWise):
    """
    Concatenate two string columns together with optional 'sep' argument.

    >>> import pandas as pd
    >>> from blaze import symbol, compute, dshape

    >>> ds = dshape('3 * {name: ?string, comment: ?string, num: int32}')
    >>> s = symbol('s', dshape=ds)
    >>> data = [('al', 'good', 0), ('suri', 'not good', 1), ('jinka', 'ok', 2)]
    >>> df = pd.DataFrame(data, columns=['name', 'comment', 'num'])

    >>> compute(s.name.str.cat(s.comment, sep=' -- '), df)
    0          al -- good
    1    suri -- not good
    2         jinka -- ok
    Name: name, dtype: object

    For rows with null entries, it returns null. This is consistent with
    default pandas behavior with kwarg: na_rep=None.

    >>> data = [(None, None, 0), ('suri', 'not good', 1), ('jinka', None, 2)]
    >>> df = pd.DataFrame(data, columns=['name', 'comment', 'num'])
    >>> compute(s.name.str.cat(s.comment, sep=' -- '), df)
    0                 NaN
    1    suri -- not good
    2                 NaN
    Name: name, dtype: object

    """
    _arguments = 'lhs', 'rhs', 'sep'
    __inputs__ = 'lhs', 'rhs'

    def _dshape(self):
        '''
        since pandas supports concat for string columns, do the same for blaze
        '''
        shape = self.lhs.dshape.shape
        if isinstance(self.lhs.schema.measure, Option):
            schema = self.lhs.schema
        elif isinstance(self.rhs.schema.measure, Option):
            schema = self.rhs.schema
        else:
            _, lhs_encoding = self.lhs.schema.measure.parameters
            _, rhs_encoding = self.rhs.schema.measure.parameters
            assert lhs_encoding == rhs_encoding
            # convert fixed length string to variable length string
            schema = DataShape(String(None, lhs_encoding))

        return DataShape(*(shape + (schema,)))


@copydoc(StrCat)
def cat(lhs, rhs, sep=None):
    """
    returns lhs + sep + rhs

    Raises:
        Invoking on a non string column raises a TypeError
        If kwarg 'sep' is not a string, raises a TypeError
    """
    # pandas supports concat for string columns only, do the same for blaze
    if not isstring(rhs.dshape):
        raise TypeError("can only concat string columns")

    _validate_optional(sep, 'sep', basestring, 'string')
    return StrCat(lhs, rhs, sep=sep)


def isstring(ds):
    measure = ds.measure
    return isinstance(getattr(measure, 'ty', measure), String)


_mod, _rmod = _mkbin('mod', Interp)
_mul, _rmul = _mkbin('mul', Repeat)


class str_ns(object):

    def __init__(self, field):
        self.field = field

    def upper(self): return upper(self.field)
    def lower(self): return lower(self.field)
    def len(self): return len(self.field)
    def like(self, pattern): return like(self.field, pattern)
    def cat(self, other, sep=None): return cat(self.field, other, sep=sep)
    def find(self, sub): return find(self.field, sub)
    def isalnum(self): return isalnum(self.field)
    def isalpha(self): return isalpha(self.field)
    def isdecimal(self): return isdecimal(self.field)
    def isdigit(self): return isdigit(self.field)
    def islower(self): return islower(self.field)
    def isnumeric(self): return isnumeric(self.field)
    def isspace(self): return isspace(self.field)
    def istitle(self): return istitle(self.field)
    def isupper(self): return isupper(self.field)
    def replace(self, old, new, max=None): return replace(self.field, old, new, max)
    def capitalize(self): return capitalize(self.field)
    def pad(self, width, side=None, fillchar=None): return pad(self.field, width, side, fillchar)
    def strip(self): return strip(self.field)
    def lstrip(self): return lstrip(self.field)
    def rstrip(self): return rstrip(self.field)
    def __getitem__(self, idx): return slice(self.field, idx)
    def slice_replace(self, start=None, stop=None, repl=None):
        return slice_replace(self.field, start, stop, repl)

class str(object):

    __name__ = 'str'

    def __get__(self, obj, type):
        return str_ns(obj) if obj is not None else self


@deprecated('0.11', replacement='len()')
def str_len(*args, **kwds): return len(*args, **kwds)
@deprecated('0.11', replacement='upper()')
def str_upper(*args, **kwds): return upper(*args, **kwds)
@deprecated('0.11', replacement='lower()')
def str_lower(*args, **kwds): return lower(*args, **kwds)
@deprecated('0.11', replacement='cat(lhs, rhs, sep=None)')
def str_cat(*args, **kwds): return cat(*args, **kwds)


schema_method_list.extend([(isstring,
                            set([_add,
                                 _radd,
                                 _mod,
                                 _rmod,
                                 _mul,
                                 _rmul,
                                 str(),
                                 repeat,
                                 interp,
                                 like,
                                 str_len, # deprecated
                                 str_upper, # deprecated
                                 str_lower, # deprecated
                                 str_cat]))]) # deprecated



from .dispatch import dispatch
from .compatibility import basestring
from blaze.interactive import _Data
from blaze.interactive import data as bz_data

@dispatch(object, (basestring, list, tuple))
def create_index(t, column_name_or_names, name=None):
    """Create an index on a column.

    Parameters
    ----------
    o : table-like
    index_name : str
        The name of the index to create
    column_name_or_names : string, list, tuple
        A column name to index on, or a list or tuple for a composite index

    Examples
    --------
    >>> # Using SQLite
    >>> from blaze import SQL
    >>> # create a table called 'tb', in memory
    >>> sql = SQL('sqlite:///:memory:', 'tb',
    ...           schema='{id: int64, value: float64, categ: string}')
    >>> dta = [(1, 2.0, 'a'), (2, 3.0, 'b'), (3, 4.0, 'c')]
    >>> sql.extend(dta)
    >>> # create an index on the 'id' column (for SQL we must provide a name)
    >>> sql.table.indexes
    set()
    >>> create_index(sql, 'id', name='id_index')
    >>> sql.table.indexes
    {Index('id_index', Column('id', BigInteger(), table=<tb>, nullable=False))}
    """
    raise NotImplementedError("create_index not implemented for type %r" %
                              type(t).__name__)

@dispatch(_Data, (basestring, list, tuple))
def create_index(dta, column_name_or_names, name=None, **kwargs):
    return create_index(dta.data, column_name_or_names, name=name, **kwargs)


@dispatch(basestring, (basestring, list, tuple))
def create_index(uri, column_name_or_names, name=None, **kwargs):
    dta = bz_data(uri, **kwargs)
    create_index(dta, column_name_or_names, name=name)
    return dta




import datetime

from blaze.compute.pyfunc import symbol, lambdify, cos, math, broadcast
from blaze.compute.pyfunc import _print_python
from blaze.expr.broadcast import broadcast_collect


t = symbol('t', '{x: int, y: int, z: int, when: datetime}')


def test_simple():
    f = lambdify([t], t.x + t.y)
    assert f((1, 2, 3, 4)) == 1 + 2

    f = lambdify([t.x, t.y], t.x + t.y)
    assert f(1, 2) == 1 + 2


def test_datetime():
    f = lambdify([t], t.x + t.when.year)

    assert f((1, 2, 3, datetime.datetime(2000, 1, 1))) == 1 + 2000


def inc(x):
    return x + 1


def test_map():
    f = lambdify([t], t.x + t.y.map(inc, 'int'))
    assert f((1, 2, 3, 4)) == 1 + inc(2)


def test_math():
    f = lambdify([t], abs(t.x) + cos(t.y))
    assert f((-1, 0, 3, 4)) == 1 + math.cos(0.0)


def test_datetime_literals_and__print_python():
    _print_python(datetime.date(2000, 1, 1)) == \
            'datetime.date(2000, 1, 1)', {'datetime': datetime}


def test_datetime_literals():
    f = lambdify([t], t.when > '2000-01-01')
    assert f((1, 0, 3, datetime.datetime(2000, 1, 2)))
    assert not f((1, 0, 3, datetime.datetime(1999, 1, 2)))


def test_broadcast_collect():
    t = symbol('t', 'var * {x: int, y: int, z: int, when: datetime}')

    expr = t.distinct()
    expr = expr.x + 2 * expr.y
    expr = expr.distinct()

    result = broadcast_collect(expr)

    expected = t.distinct()
    expected = broadcast(expected.x + 2 * expected.y, [expected])
    expected = expected.distinct()

    assert result.isidentical(expected)


def test_pyfunc_works_with_invalid_python_names():
    x = symbol('x-y.z', 'int')
    f = lambdify([x], x + 1)
    assert f(1) == 2

    t = symbol('t', '{"x.y": int, "y z": int}')
    f = lambdify([t], t.x_y + t.y_z)
    assert f((1, 2)) == 3


def test_usub():
    x = symbol('x', 'float64')
    f = lambdify([x], -x)
    assert f(1.0) == -1.0


def test_not():
    x = symbol('x', 'bool')
    f = lambdify([x], ~x)
    r = f(True)
    assert isinstance(r, bool) and not r

    r = f(False)
    assert isinstance(r, bool) and r



from __future__ import absolute_import, division, print_function

import itertools

from datashape import DataShape, iscollection
from toolz import curry, concat

from .collections import Shift
from .core import Node
from .expressions import Field, Map, ElemWise, symbol, shape, Coerce
from .arithmetic import maxshape, UnaryOp, BinOp
from .strings import Like, StrCat
from .datetime import DateTime

__all__ = ['broadcast', 'Broadcast', 'scalar_symbols', 'broadcast_collect']


def broadcast(expr, leaves, scalars=None):
    scalars = scalars or scalar_symbols(leaves)
    assert len(scalars) == len(leaves)
    return Broadcast(tuple(leaves),
                     tuple(scalars),
                     expr._subs(dict(zip(leaves, scalars))))


class Broadcast(ElemWise):

    """ Fuse scalar expressions over collections

    Given elementwise operations on collections, e.g.

    >>> from blaze import sin
    >>> a = symbol('a', '100 * int')
    >>> t = symbol('t', '100 * {x: int, y: int}')

    >>> expr = sin(a) + t.y**2

    It may be best to represent this as a scalar expression mapped over a
    collection

    >>> sa = symbol('a', 'int')
    >>> st = symbol('t', '{x: int, y: int}')

    >>> sexpr = sin(sa) + st.y**2

    >>> expr = Broadcast((a, t), (sa, st), sexpr)

    This provides opportunities for optimized computation.

    In practice, expressions are often collected into Broadcast expressions
    automatically.  This class is mainly intented for internal use.
    """
    _arguments = '_children', '_scalars', '_scalar_expr'

    def _dshape(self):
        myshape = maxshape(map(shape, self._children))
        return DataShape(*(myshape + (self._scalar_expr.schema,)))

    @property
    def _inputs(self):
        return self._children

    @property
    def _name(self):
        return self._scalar_expr._name

    @property
    def _full_expr(self):
        return self._scalar_expr._subs(
            dict(zip(self._scalars, self._children))
        )

    def _traverse(self):
        for item in super(Broadcast, self)._traverse():
            yield item
        # also yield all the items in the traversal of our full expression
        # without yielding the full expression itself.
        for item in itertools.islice(self._full_expr._traverse(), 1, None):
            yield item


def scalar_symbols(exprs):
    """
    Gives a sequence of scalar symbols to mirror these expressions

    Examples
    --------

    >>> x = symbol('x', '5 * 3 * int32')
    >>> y = symbol('y', '5 * 3 * int32')

    >>> xx, yy = scalar_symbols([x, y])

    >>> xx._name, xx.dshape
    ('x', dshape("int32"))
    >>> yy._name, yy.dshape
    ('y', dshape("int32"))
    """
    new_names = ('_%d' % i for i in itertools.count(1))

    scalars = []
    names = set()
    for expr in exprs:
        if expr._name and expr._name not in names:
            name = expr._name
            names.add(name)
        else:
            name = next(new_names)

        s = symbol(name, expr.schema)
        scalars.append(s)
    return scalars


Broadcastable = (Map, Field, DateTime, UnaryOp, BinOp, Coerce, Shift, Like, StrCat)
WantToBroadcast = (Map, DateTime, UnaryOp, BinOp, Coerce, Shift, Like, StrCat)


def broadcast_collect(expr,
                      broadcastable=Broadcastable,
                      want_to_broadcast=WantToBroadcast,
                      no_recurse=None):
    """ Collapse expression down using Broadcast - Tabular cases only

    Expressions of type Broadcastables are swallowed into Broadcast
    operations

    >>> t = symbol('t', 'var * {x: int, y: int, z: int, when: datetime}')
    >>> expr = (t.x + 2*t.y).distinct()

    >>> broadcast_collect(expr)
    distinct(Broadcast(_children=(t,), _scalars=(t,), _scalar_expr=t.x + (2 * t.y)))

    >>> from blaze import exp
    >>> expr = t.x + 2 * exp(-(t.x - 1.3) ** 2)
    >>> broadcast_collect(expr)
    Broadcast(_children=(t,), _scalars=(t,), _scalar_expr=t.x + (2 * (exp(-((t.x - 1.3) ** 2)))))
    """
    if (isinstance(expr, want_to_broadcast) and
            iscollection(expr.dshape)):
        leaves = leaves_of_type(broadcastable, expr)
        expr = broadcast(expr, sorted(leaves, key=str))

    if no_recurse is not None and isinstance(expr, no_recurse):
        return expr

    # Recurse down
    children = (
        broadcast_collect(i, broadcastable, want_to_broadcast, no_recurse)
        for i in expr._inputs
    )
    return expr._subs({e: c for e, c in zip(expr._inputs, children)})


@curry
def leaves_of_type(types, expr):
    """ Leaves of an expression skipping all operations of type ``types``
    """
    if not isinstance(expr, types):
        return set([expr])
    else:
        return set.union(*map(leaves_of_type(types), expr._inputs))



from __future__ import absolute_import, division, print_function

import os
import pytest
tb = pytest.importorskip('tables')

import numpy as np

from blaze.compute.core import compute
from blaze.expr import symbol
from blaze import drop, discover, create_index
from blaze.utils import tmpfile


t = symbol('t', 'var * {id: int, name: string, amount: int}')

x = np.array([(1, 'Alice', 100),
              (2, 'Bob', -200),
              (3, 'Charlie', 300),
              (4, 'Denis', 400),
              (5, 'Edith', -500)],
             dtype=[('id', '<i8'), ('name', 'S7'), ('amount', '<i8')])


@pytest.yield_fixture
def data():
    with tmpfile('.h5') as filename:
        f = tb.open_file(filename, mode='w')
        d = f.create_table('/', 'title',  x)
        yield d
        d.close()
        f.close()


@pytest.yield_fixture
def csi_data():
    with tmpfile('.h5') as filename:
        f = tb.open_file(filename, mode='w')
        d = f.create_table('/', 'title', x)
        d.cols.amount.create_csindex()
        d.cols.id.create_csindex()
        yield d
        d.close()
        f.close()


@pytest.yield_fixture
def idx_data():
    with tmpfile('.h5') as fn:
        f = tb.open_file(fn, mode='w')
        d = f.create_table('/', 'title', x)
        d.cols.amount.create_index()
        d.cols.id.create_index()
        yield d
        d.close()
        f.close()


def eq(a, b):
    return (a == b).all()


def test_discover_datashape(data):
    ds = discover(data)
    t = symbol('t', ds)
    columns = t.fields
    assert columns is not None


def test_symbol(data):
    assert compute(t, data) == data
    assert isinstance(data, tb.Table)


def test_single_column(data):
    assert eq(compute(t['name'], data), x['name'])


def test_projection(data):
    assert eq(compute(t[['name', 'amount']], data), x[['name', 'amount']])


def test_eq(data):
    assert eq(compute(t['amount'] == 100, data), x['amount'] == 100)


def test_scalar_ops(data):
    from operator import add, sub, mul, truediv

    for op in (add, sub, mul, truediv):
        assert eq(compute(op(t.amount, 10), data), op(x['amount'], 10))
        assert eq(compute(op(t.amount, t.id), data), op(x['amount'], x['id']))
        assert eq(compute(op(10.0, t.amount), data), op(10.0, x['amount']))
        assert eq(compute(op(10, t.amount), data), op(10, x['amount']))


def test_neg(data):
    assert eq(compute(-t.amount, data), -x['amount'])


def test_failing_floordiv(data):
    from operator import floordiv as op

    with pytest.raises(TypeError):
        assert eq(compute(op(t.amount, 10), data), op(x['amount'], 10))

    with pytest.raises(TypeError):
        assert eq(compute(op(t.amount, t.id), data), op(x['amount'], x['id']))

    with pytest.raises(TypeError):
        assert eq(compute(op(10.0, t.amount), data), op(10.0, x['amount']))

    with pytest.raises(TypeError):
        assert eq(compute(op(10, t.amount), data), op(10, x['amount']))


def test_selection(data):
    assert eq(compute(t[t['amount'] == 100], data), x[x['amount'] == 0])
    assert eq(compute(t[t['amount'] < 0], data), x[x['amount'] < 0])


def test_arithmetic(data):
    assert eq(compute(t['amount'] + t['id'], data), x['amount'] + x['id'])
    assert eq(compute(t['amount'] * t['id'], data), x['amount'] * x['id'])
    assert eq(compute(t['amount'] % t['id'], data), x['amount'] % x['id'])
    assert eq(compute(t['amount'] + t['id'] + 3, data),
              x['amount'] + x['id'] + 3)


def test_reductions(data):
    assert compute(t['amount'].count(), data) == len(x['amount'])
    assert compute(t['amount'].sum(), data) == x['amount'].sum()
    assert compute(t['amount'].mean(), data) == x['amount'].mean()
    assert compute(t.amount[0], data) == x['amount'][0]
    assert compute(t.amount[-1], data) == x['amount'][-1]


class TestTopLevelReductions(object):
    def test_count(self, data):
        from blaze import count
        assert compute(count(t['amount']), data) == len(x['amount'])

    def test_sum(self, data):
        from blaze import sum
        assert compute(sum(t['amount']), data) == x['amount'].sum()

    def test_mean(self, data):
        from blaze import mean
        assert compute(mean(t['amount']), data) == x['amount'].mean()


class TestFailingSort(object):
    """These fail because we haven't created a completely sorted index"""

    def test_basic(self, data):
        with pytest.raises(ValueError):
            compute(t.sort('id'), data)

    @pytest.mark.xfail(
        raises=TypeError,
        reason='PyTables does not support multiple column sorting'
    )
    def test_multiple_columns(self, data):
        compute(t.sort(['amount', 'id']), data)

    @pytest.mark.xfail(
        raises=TypeError,
        reason='PyTables does not support multiple column sorting'
    )
    def test_multiple_columns_sorted_data(self, csi_data):
        compute(t.sort(['amount', 'id']), csi_data)


class TestCSISort(object):
    def test_basic(self, csi_data):
        assert eq(compute(t.sort('amount'), csi_data),
                  np.sort(x, order='amount'))
        assert eq(compute(t.sort('id'), csi_data),
                  np.sort(x, order='id'))

    def test_column_expr(self, csi_data):
        assert eq(compute(t.sort(t.amount), csi_data),
                  np.sort(x, order='amount'))
        assert eq(compute(t.sort(t.id), csi_data),
                  np.sort(x, order='id'))

    def test_non_existent_column(self, csi_data):
        with pytest.raises(ValueError):
            compute(t.sort('not here'), csi_data)

    def test_ascending(self, csi_data):
        assert eq(compute(t.sort('amount', ascending=False), csi_data),
                  np.sort(x, order='amount')[::-1])
        assert eq(compute(t.sort('amount', ascending=False), csi_data),
                  np.sort(x, order='amount')[::-1])


class TestIndexSort(object):
    """Fails with a partially sorted index"""

    @pytest.mark.xfail(
        raises=ValueError,
        reason='PyTables cannot sort with a standard index'
    )
    def test_basic(self, idx_data):
        compute(t.sort('amount'), idx_data)

    @pytest.mark.xfail(
        raises=ValueError,
        reason='PyTables cannot sort with a standard index'
    )
    def test_ascending(self, idx_data):
        compute(t.sort('amount', ascending=False), idx_data)


def test_head(data):
    assert eq(compute(t.head(2), data), x[:2])
    assert eq(compute(t.amount.head(2), data), x['amount'][:2])


@pytest.yield_fixture
def pyt():
    tb = pytest.importorskip('tables')
    fn = 'test.pyt.h5'
    f = tb.open_file(fn, mode='w')
    d = f.create_table('/', 'test', x)
    try:
        yield d
    finally:
        d.close()
        f.close()
        try:
            os.remove(fn)
        except OSError:
            pass


def test_drop(pyt):
    drop(pyt)
    with pytest.raises(tb.ClosedNodeError):
        drop(pyt)


def test_create_index(pyt):
    create_index(pyt, 'id')
    assert 'id' in pyt.colindexes


def test_create_multiple_indexes(pyt):
    create_index(pyt, ['id', 'amount'])
    assert len(pyt.colindexes) == 2
    assert 'id' in pyt.colindexes
    assert 'amount' in pyt.colindexes


def test_create_multiple_indexes_fails(pyt):
    with pytest.raises(ValueError):
        create_index(pyt, ['id', 'blarg'])

    with pytest.raises(ValueError):
        create_index(pyt, ['foo', 'bar'])


def test_create_index_fails(pyt):
    with pytest.raises(AttributeError):
        create_index(pyt, 'no column here!')


def test_nrows():
    assert compute(t.nrows, x) == len(x)


def test_nelements():
    assert compute(t.nelements(axis=0), x) == len(x)
    assert compute(t.nelements(), x) == len(x)
